# This file is part of beets.
# Copyright 2016, Adrian Sampson.
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.

"""Tests for the album art fetchers."""

from __future__ import annotations

import os
import shutil
import unittest
from pathlib import Path
from typing import TYPE_CHECKING
from unittest.mock import patch

import confuse
import pytest
import responses

from beets import config, importer, logging, util
from beets.autotag.distance import Distance
from beets.autotag.hooks import AlbumInfo, AlbumMatch
from beets.test import _common
from beets.test.helper import (
    BeetsTestCase,
    CleanupModulesMixin,
    FetchImageHelper,
    capture_log,
)
from beets.util import syspath
from beets.util.artresizer import ArtResizer
from beetsplug import fetchart

logger = logging.getLogger("beets.test_art")

if TYPE_CHECKING:
    from collections.abc import Iterator, Sequence

    from beets.library import Album


class Settings:
    """Used to pass settings to the ArtSources when the plugin isn't fully
    instantiated.
    """

    def __init__(self, **kwargs):
        for k, v in kwargs.items():
            setattr(self, k, v)


class DummyRemoteArtSource(fetchart.RemoteArtSource):
    NAME = "Dummy Art Source"
    ID = "dummy"

    def get(
        self,
        album: Album,
        plugin: fetchart.FetchArtPlugin,
        paths: None | Sequence[bytes],
    ) -> Iterator[fetchart.Candidate]:
        return iter(())


class UseThePlugin(CleanupModulesMixin, BeetsTestCase):
    modules = (fetchart.__name__, ArtResizer.__module__)

    def setUp(self):
        super().setUp()
        self.plugin = fetchart.FetchArtPlugin()


class FetchImageTestCase(FetchImageHelper, UseThePlugin):
    pass


class CAAHelper:
    """Helper mixin for mocking requests to the Cover Art Archive."""

    MBID_RELASE = "rid"
    MBID_GROUP = "rgid"

    RELEASE_URL = f"coverartarchive.org/release/{MBID_RELASE}"
    GROUP_URL = f"coverartarchive.org/release-group/{MBID_GROUP}"

    RELEASE_URL = f"https://{RELEASE_URL}"
    GROUP_URL = f"https://{GROUP_URL}"

    RESPONSE_RELEASE = """{
    "images": [
      {
        "approved": false,
        "back": false,
        "comment": "GIF",
        "edit": 12345,
        "front": true,
        "id": 12345,
        "image": "http://coverartarchive.org/release/rid/12345.gif",
        "thumbnails": {
          "1200": "http://coverartarchive.org/release/rid/12345-1200.jpg",
          "250": "http://coverartarchive.org/release/rid/12345-250.jpg",
          "500": "http://coverartarchive.org/release/rid/12345-500.jpg",
          "large": "http://coverartarchive.org/release/rid/12345-500.jpg",
          "small": "http://coverartarchive.org/release/rid/12345-250.jpg"
        },
        "types": [
          "Front"
        ]
      },
      {
        "approved": false,
        "back": false,
        "comment": "",
        "edit": 12345,
        "front": false,
        "id": 12345,
        "image": "http://coverartarchive.org/release/rid/12345.jpg",
        "thumbnails": {
          "1200": "http://coverartarchive.org/release/rid/12345-1200.jpg",
          "250": "http://coverartarchive.org/release/rid/12345-250.jpg",
          "500": "http://coverartarchive.org/release/rid/12345-500.jpg",
          "large": "http://coverartarchive.org/release/rid/12345-500.jpg",
          "small": "http://coverartarchive.org/release/rid/12345-250.jpg"
        },
        "types": [
          "Front"
        ]
      }
    ],
    "release": "https://musicbrainz.org/release/releaseid"
}"""
    RESPONSE_RELEASE_WITHOUT_THUMBNAILS = """{
    "images": [
      {
        "approved": false,
        "back": false,
        "comment": "GIF",
        "edit": 12345,
        "front": true,
        "id": 12345,
        "image": "http://coverartarchive.org/release/rid/12345.gif",
        "types": [
          "Front"
        ]
      },
      {
        "approved": false,
        "back": false,
        "comment": "",
        "edit": 12345,
        "front": false,
        "id": 12345,
        "image": "http://coverartarchive.org/release/rid/12345.jpg",
        "thumbnails": {
            "large": "http://coverartarchive.org/release/rgid/12345-500.jpg",
            "small": "http://coverartarchive.org/release/rgid/12345-250.jpg"
        },
        "types": [
          "Front"
        ]
      }
    ],
    "release": "https://musicbrainz.org/release/releaseid"
}"""
    RESPONSE_GROUP = """{
        "images": [
          {
            "approved": false,
            "back": false,
            "comment": "",
            "edit": 12345,
            "front": true,
            "id": 12345,
            "image": "http://coverartarchive.org/release/releaseid/12345.jpg",
            "thumbnails": {
              "1200": "http://coverartarchive.org/release/rgid/12345-1200.jpg",
              "250": "http://coverartarchive.org/release/rgid/12345-250.jpg",
              "500": "http://coverartarchive.org/release/rgid/12345-500.jpg",
              "large": "http://coverartarchive.org/release/rgid/12345-500.jpg",
              "small": "http://coverartarchive.org/release/rgid/12345-250.jpg"
            },
            "types": [
              "Front"
            ]
          }
        ],
        "release": "https://musicbrainz.org/release/release-id"
    }"""
    RESPONSE_GROUP_WITHOUT_THUMBNAILS = """{
        "images": [
          {
            "approved": false,
            "back": false,
            "comment": "",
            "edit": 12345,
            "front": true,
            "id": 12345,
            "image": "http://coverartarchive.org/release/releaseid/12345.jpg",
            "types": [
              "Front"
            ]
          }
        ],
        "release": "https://musicbrainz.org/release/release-id"
    }"""

    def mock_caa_response(self, url, json):
        responses.add(
            responses.GET, url, body=json, content_type="application/json"
        )


class FetchImageTest(FetchImageTestCase):
    URL = "http://example.com/test.jpg"

    def setUp(self):
        super().setUp()
        self.dpath = os.path.join(self.temp_dir, b"arttest")
        self.source = DummyRemoteArtSource(logger, self.plugin.config)
        self.settings = Settings(maxwidth=0)
        self.candidate = fetchart.Candidate(
            logger, self.source.ID, url=self.URL
        )

    def test_invalid_type_returns_none(self):
        self.mock_response(self.URL, "image/watercolour")
        self.source.fetch_image(self.candidate, self.settings)
        assert self.candidate.path is None

    def test_jpeg_type_returns_path(self):
        self.mock_response(self.URL, "image/jpeg")
        self.source.fetch_image(self.candidate, self.settings)
        assert self.candidate.path is not None

    def test_extension_set_by_content_type(self):
        self.mock_response(self.URL, "image/png")
        self.source.fetch_image(self.candidate, self.settings)
        assert os.path.splitext(self.candidate.path)[1] == b".png"
        assert Path(os.fsdecode(self.candidate.path)).exists()

    def test_does_not_rely_on_server_content_type(self):
        self.mock_response(self.URL, "image/jpeg", "image/png")
        self.source.fetch_image(self.candidate, self.settings)
        assert os.path.splitext(self.candidate.path)[1] == b".png"
        assert Path(os.fsdecode(self.candidate.path)).exists()


class FSArtTest(UseThePlugin):
    def setUp(self):
        super().setUp()
        self.dpath = os.path.join(self.temp_dir, b"arttest")
        os.mkdir(syspath(self.dpath))

        self.source = fetchart.FileSystem(logger, self.plugin.config)
        self.settings = Settings(
            cautious=False, cover_names=("art",), fallback=None
        )

    def test_finds_jpg_in_directory(self):
        _common.touch(os.path.join(self.dpath, b"a.jpg"))
        candidate = next(self.source.get(None, self.settings, [self.dpath]))
        assert candidate.path == os.path.join(self.dpath, b"a.jpg")

    def test_appropriately_named_file_takes_precedence(self):
        _common.touch(os.path.join(self.dpath, b"a.jpg"))
        _common.touch(os.path.join(self.dpath, b"art.jpg"))
        candidate = next(self.source.get(None, self.settings, [self.dpath]))
        assert candidate.path == os.path.join(self.dpath, b"art.jpg")

    def test_non_image_file_not_identified(self):
        _common.touch(os.path.join(self.dpath, b"a.txt"))
        with pytest.raises(StopIteration):
            next(self.source.get(None, self.settings, [self.dpath]))

    def test_cautious_skips_fallback(self):
        _common.touch(os.path.join(self.dpath, b"a.jpg"))
        self.settings.cautious = True
        with pytest.raises(StopIteration):
            next(self.source.get(None, self.settings, [self.dpath]))

    def test_configured_fallback_is_used(self):
        fallback = os.path.join(self.temp_dir, b"a.jpg")
        _common.touch(fallback)
        self.settings.fallback = fallback
        candidate = next(self.source.get(None, self.settings, [self.dpath]))
        assert candidate.path == fallback

    def test_empty_dir(self):
        with pytest.raises(StopIteration):
            next(self.source.get(None, self.settings, [self.dpath]))

    def test_precedence_amongst_correct_files(self):
        images = [b"front-cover.jpg", b"front.jpg", b"back.jpg"]
        paths = [os.path.join(self.dpath, i) for i in images]
        for p in paths:
            _common.touch(p)
        self.settings.cover_names = ["cover", "front", "back"]
        candidates = [
            candidate.path
            for candidate in self.source.get(None, self.settings, [self.dpath])
        ]
        assert candidates == paths

    @patch("os.path.samefile")
    def test_is_candidate_fallback_os_error(self, mock_samefile):
        mock_samefile.side_effect = OSError("os error")
        fallback = os.path.join(self.temp_dir, b"a.jpg")
        self.plugin.fallback = fallback
        candidate = fetchart.Candidate(logger, self.source.ID, fallback)
        result = self.plugin._is_candidate_fallback(candidate)
        mock_samefile.assert_called_once()
        assert not result


class CombinedTest(FetchImageTestCase, CAAHelper):
    ASIN = "xxxx"
    MBID = "releaseid"
    AMAZON_URL = f"https://images.amazon.com/images/P/{ASIN}.01.LZZZZZZZ.jpg"
    AAO_URL = f"https://www.albumart.org/index_detail.php?asin={ASIN}"

    def setUp(self):
        super().setUp()
        self.dpath = os.path.join(self.temp_dir, b"arttest")
        os.mkdir(syspath(self.dpath))

    def test_main_interface_returns_amazon_art(self):
        self.mock_response(self.AMAZON_URL)
        album = _common.Bag(asin=self.ASIN)
        candidate = self.plugin.art_for_album(album, None)
        assert candidate is not None

    def test_main_interface_returns_none_for_missing_asin_and_path(self):
        album = _common.Bag()
        candidate = self.plugin.art_for_album(album, None)
        assert candidate is None

    def test_main_interface_gives_precedence_to_fs_art(self):
        _common.touch(os.path.join(self.dpath, b"art.jpg"))
        self.mock_response(self.AMAZON_URL)
        album = _common.Bag(asin=self.ASIN)
        candidate = self.plugin.art_for_album(album, [self.dpath])
        assert candidate is not None
        assert candidate.path == os.path.join(self.dpath, b"art.jpg")

    def test_main_interface_falls_back_to_amazon(self):
        self.mock_response(self.AMAZON_URL)
        album = _common.Bag(asin=self.ASIN)
        candidate = self.plugin.art_for_album(album, [self.dpath])
        assert candidate is not None
        assert not candidate.path.startswith(self.dpath)

    def test_main_interface_tries_amazon_before_aao(self):
        self.mock_response(self.AMAZON_URL)
        album = _common.Bag(asin=self.ASIN)
        self.plugin.art_for_album(album, [self.dpath])
        assert len(responses.calls) == 1
        assert responses.calls[0].request.url == self.AMAZON_URL

    def test_main_interface_falls_back_to_aao(self):
        self.mock_response(self.AMAZON_URL, content_type="text/html")
        album = _common.Bag(asin=self.ASIN)
        self.plugin.art_for_album(album, [self.dpath])
        assert responses.calls[-1].request.url == self.AAO_URL

    def test_main_interface_uses_caa_when_mbid_available(self):
        self.mock_caa_response(self.RELEASE_URL, self.RESPONSE_RELEASE)
        self.mock_caa_response(self.GROUP_URL, self.RESPONSE_GROUP)
        self.mock_response(
            "http://coverartarchive.org/release/rid/12345.gif",
            content_type="image/gif",
        )
        self.mock_response(
            "http://coverartarchive.org/release/rid/12345.jpg",
            content_type="image/jpeg",
        )
        album = _common.Bag(
            mb_albumid=self.MBID_RELASE,
            mb_releasegroupid=self.MBID_GROUP,
            asin=self.ASIN,
        )
        candidate = self.plugin.art_for_album(album, None)
        assert candidate is not None
        assert len(responses.calls) == 3
        assert responses.calls[0].request.url == self.RELEASE_URL

    def test_local_only_does_not_access_network(self):
        album = _common.Bag(mb_albumid=self.MBID, asin=self.ASIN)
        self.plugin.art_for_album(album, None, local_only=True)
        assert len(responses.calls) == 0

    def test_local_only_gets_fs_image(self):
        _common.touch(os.path.join(self.dpath, b"art.jpg"))
        album = _common.Bag(mb_albumid=self.MBID, asin=self.ASIN)
        candidate = self.plugin.art_for_album(
            album, [self.dpath], local_only=True
        )
        assert candidate is not None
        assert candidate.path == os.path.join(self.dpath, b"art.jpg")
        assert len(responses.calls) == 0


class AAOTest(UseThePlugin):
    ASIN = "xxxx"
    AAO_URL = f"https://www.albumart.org/index_detail.php?asin={ASIN}"

    def setUp(self):
        super().setUp()
        self.source = fetchart.AlbumArtOrg(logger, self.plugin.config)
        self.settings = Settings()

    @responses.activate
    def run(self, *args, **kwargs):
        super().run(*args, **kwargs)

    def mock_response(self, url, body):
        responses.add(responses.GET, url, body=body, content_type="text/html")

    def test_aao_scraper_finds_image(self):
        body = """
        <br />
        <a href=\"TARGET_URL\" title=\"View larger image\"
           class=\"thickbox\" style=\"color: #7E9DA2; text-decoration:none;\">
        <img src=\"http://www.albumart.org/images/zoom-icon.jpg\"
       alt=\"View larger image\" width=\"17\" height=\"15\"  border=\"0\"/></a>
        """
        self.mock_response(self.AAO_URL, body)
        album = _common.Bag(asin=self.ASIN)
        candidate = next(self.source.get(album, self.settings, []))
        assert candidate.url == "TARGET_URL"

    def test_aao_scraper_returns_no_result_when_no_image_present(self):
        self.mock_response(self.AAO_URL, "blah blah")
        album = _common.Bag(asin=self.ASIN)
        with pytest.raises(StopIteration):
            next(self.source.get(album, self.settings, []))


class ITunesStoreTest(UseThePlugin):
    def setUp(self):
        super().setUp()
        self.source = fetchart.ITunesStore(logger, self.plugin.config)
        self.settings = Settings()
        self.album = _common.Bag(albumartist="some artist", album="some album")

    @responses.activate
    def run(self, *args, **kwargs):
        super().run(*args, **kwargs)

    def mock_response(self, url, json):
        responses.add(
            responses.GET, url, body=json, content_type="application/json"
        )

    def test_itunesstore_finds_image(self):
        json = """{
                    "results":
                        [
                            {
                                "artistName": "some artist",
                                "collectionName": "some album",
                                "artworkUrl100": "url_to_the_image"
                            }
                        ]
                  }"""
        self.mock_response(fetchart.ITunesStore.API_URL, json)
        candidate = next(self.source.get(self.album, self.settings, []))
        assert candidate.url == "url_to_the_image"
        assert candidate.match == fetchart.MetadataMatch.EXACT

    def test_itunesstore_no_result(self):
        json = '{"results": []}'
        self.mock_response(fetchart.ITunesStore.API_URL, json)
        expected = "got no results"

        with capture_log("beets.test_art") as logs:
            with pytest.raises(StopIteration):
                next(self.source.get(self.album, self.settings, []))
        assert expected in logs[1]

    def test_itunesstore_requestexception(self):
        responses.add(
            responses.GET,
            fetchart.ITunesStore.API_URL,
            json={"error": "not found"},
            status=404,
        )
        expected = "iTunes search failed: 404 Client Error"

        with capture_log("beets.test_art") as logs:
            with pytest.raises(StopIteration):
                next(self.source.get(self.album, self.settings, []))
        assert expected in logs[1]

    def test_itunesstore_fallback_match(self):
        json = """{
                    "results":
                        [
                            {
                                "collectionName": "some album",
                                "artworkUrl100": "url_to_the_image"
                            }
                        ]
                  }"""
        self.mock_response(fetchart.ITunesStore.API_URL, json)
        candidate = next(self.source.get(self.album, self.settings, []))
        assert candidate.url == "url_to_the_image"
        assert candidate.match == fetchart.MetadataMatch.FALLBACK

    def test_itunesstore_returns_result_without_artwork(self):
        json = """{
                    "results":
                        [
                            {
                                "artistName": "some artist",
                                "collectionName": "some album"
                            }
                        ]
                  }"""
        self.mock_response(fetchart.ITunesStore.API_URL, json)
        expected = "Malformed itunes candidate"

        with capture_log("beets.test_art") as logs:
            with pytest.raises(StopIteration):
                next(self.source.get(self.album, self.settings, []))
        assert expected in logs[1]

    def test_itunesstore_returns_no_result_when_error_received(self):
        json = '{"error": {"errors": [{"reason": "some reason"}]}}'
        self.mock_response(fetchart.ITunesStore.API_URL, json)
        expected = "not found in json. Fields are"

        with capture_log("beets.test_art") as logs:
            with pytest.raises(StopIteration):
                next(self.source.get(self.album, self.settings, []))
        assert expected in logs[1]

    def test_itunesstore_returns_no_result_with_malformed_response(self):
        json = """bla blup"""
        self.mock_response(fetchart.ITunesStore.API_URL, json)
        expected = "Could not decode json response:"

        with capture_log("beets.test_art") as logs:
            with pytest.raises(StopIteration):
                next(self.source.get(self.album, self.settings, []))
        assert expected in logs[1]


class GoogleImageTest(UseThePlugin):
    def setUp(self):
        super().setUp()
        self.source = fetchart.GoogleImages(logger, self.plugin.config)
        self.settings = Settings()

    @responses.activate
    def run(self, *args, **kwargs):
        super().run(*args, **kwargs)

    def mock_response(self, url, json):
        responses.add(
            responses.GET, url, body=json, content_type="application/json"
        )

    def test_google_art_finds_image(self):
        album = _common.Bag(albumartist="some artist", album="some album")
        json = '{"items": [{"link": "url_to_the_image"}]}'
        self.mock_response(fetchart.GoogleImages.URL, json)
        candidate = next(self.source.get(album, self.settings, []))
        assert candidate.url == "url_to_the_image"

    def test_google_art_returns_no_result_when_error_received(self):
        album = _common.Bag(albumartist="some artist", album="some album")
        json = '{"error": {"errors": [{"reason": "some reason"}]}}'
        self.mock_response(fetchart.GoogleImages.URL, json)
        with pytest.raises(StopIteration):
            next(self.source.get(album, self.settings, []))

    def test_google_art_returns_no_result_with_malformed_response(self):
        album = _common.Bag(albumartist="some artist", album="some album")
        json = """bla blup"""
        self.mock_response(fetchart.GoogleImages.URL, json)
        with pytest.raises(StopIteration):
            next(self.source.get(album, self.settings, []))


class CoverArtArchiveTest(UseThePlugin, CAAHelper):
    def setUp(self):
        super().setUp()
        self.source = fetchart.CoverArtArchive(logger, self.plugin.config)
        self.settings = Settings(maxwidth=0)

    @responses.activate
    def run(self, *args, **kwargs):
        super().run(*args, **kwargs)

    def test_caa_finds_image(self):
        album = _common.Bag(
            mb_albumid=self.MBID_RELASE, mb_releasegroupid=self.MBID_GROUP
        )
        self.mock_caa_response(self.RELEASE_URL, self.RESPONSE_RELEASE)
        self.mock_caa_response(self.GROUP_URL, self.RESPONSE_GROUP)
        candidates = list(self.source.get(album, self.settings, []))
        assert len(candidates) == 3
        assert len(responses.calls) == 2
        assert responses.calls[0].request.url == self.RELEASE_URL

    def test_fetchart_uses_caa_pre_sized_maxwidth_thumbs(self):
        # CAA provides pre-sized thumbnails of width 250px, 500px, and 1200px
        # We only test with one of them here
        maxwidth = 1200
        self.settings = Settings(maxwidth=maxwidth)

        album = _common.Bag(
            mb_albumid=self.MBID_RELASE, mb_releasegroupid=self.MBID_GROUP
        )
        self.mock_caa_response(self.RELEASE_URL, self.RESPONSE_RELEASE)
        self.mock_caa_response(self.GROUP_URL, self.RESPONSE_GROUP)
        candidates = list(self.source.get(album, self.settings, []))
        assert len(candidates) == 3
        for candidate in candidates:
            assert f"-{maxwidth}.jpg" in candidate.url

    def test_caa_finds_image_if_maxwidth_is_set_and_thumbnails_is_empty(self):
        # CAA provides pre-sized thumbnails of width 250px, 500px, and 1200px
        # We only test with one of them here
        maxwidth = 1200
        self.settings = Settings(maxwidth=maxwidth)

        album = _common.Bag(
            mb_albumid=self.MBID_RELASE, mb_releasegroupid=self.MBID_GROUP
        )
        self.mock_caa_response(
            self.RELEASE_URL, self.RESPONSE_RELEASE_WITHOUT_THUMBNAILS
        )
        self.mock_caa_response(
            self.GROUP_URL,
            self.RESPONSE_GROUP_WITHOUT_THUMBNAILS,
        )
        candidates = list(self.source.get(album, self.settings, []))
        assert len(candidates) == 3
        for candidate in candidates:
            assert f"-{maxwidth}.jpg" not in candidate.url


class FanartTVTest(UseThePlugin):
    RESPONSE_MULTIPLE = """{
        "name": "artistname",
        "mbid_id": "artistid",
        "albums": {
            "thereleasegroupid": {
                "albumcover": [
                    {
                        "id": "24",
                        "url": "http://example.com/1.jpg",
                        "likes": "0"
                    },
                    {
                        "id": "42",
                        "url": "http://example.com/2.jpg",
                        "likes": "0"
                    },
                    {
                        "id": "23",
                        "url": "http://example.com/3.jpg",
                        "likes": "0"
                    }
                ],
                "cdart": [
                    {
                        "id": "123",
                        "url": "http://example.com/4.jpg",
                        "likes": "0",
                        "disc": "1",
                        "size": "1000"
                    }
                ]
            }
        }
    }"""
    RESPONSE_NO_ART = """{
        "name": "artistname",
        "mbid_id": "artistid",
        "albums": {
            "thereleasegroupid": {
               "cdart": [
                    {
                        "id": "123",
                        "url": "http://example.com/4.jpg",
                        "likes": "0",
                        "disc": "1",
                        "size": "1000"
                    }
                ]
            }
        }
    }"""
    RESPONSE_ERROR = """{
        "status": "error",
        "error message": "the error message"
    }"""
    RESPONSE_MALFORMED = "bla blup"

    def setUp(self):
        super().setUp()
        self.source = fetchart.FanartTV(logger, self.plugin.config)
        self.settings = Settings()

    @responses.activate
    def run(self, *args, **kwargs):
        super().run(*args, **kwargs)

    def mock_response(self, url, json):
        responses.add(
            responses.GET, url, body=json, content_type="application/json"
        )

    def test_fanarttv_finds_image(self):
        album = _common.Bag(mb_releasegroupid="thereleasegroupid")
        self.mock_response(
            f"{fetchart.FanartTV.API_ALBUMS}thereleasegroupid",
            self.RESPONSE_MULTIPLE,
        )
        candidate = next(self.source.get(album, self.settings, []))
        assert candidate.url == "http://example.com/1.jpg"

    def test_fanarttv_returns_no_result_when_error_received(self):
        album = _common.Bag(mb_releasegroupid="thereleasegroupid")
        self.mock_response(
            f"{fetchart.FanartTV.API_ALBUMS}thereleasegroupid",
            self.RESPONSE_ERROR,
        )
        with pytest.raises(StopIteration):
            next(self.source.get(album, self.settings, []))

    def test_fanarttv_returns_no_result_with_malformed_response(self):
        album = _common.Bag(mb_releasegroupid="thereleasegroupid")
        self.mock_response(
            f"{fetchart.FanartTV.API_ALBUMS}thereleasegroupid",
            self.RESPONSE_MALFORMED,
        )
        with pytest.raises(StopIteration):
            next(self.source.get(album, self.settings, []))

    def test_fanarttv_only_other_images(self):
        # The source used to fail when there were images present, but no cover
        album = _common.Bag(mb_releasegroupid="thereleasegroupid")
        self.mock_response(
            f"{fetchart.FanartTV.API_ALBUMS}thereleasegroupid",
            self.RESPONSE_NO_ART,
        )
        with pytest.raises(StopIteration):
            next(self.source.get(album, self.settings, []))


class ArtImporterTest(UseThePlugin):
    def setUp(self):
        super().setUp()

        # Mock the album art fetcher to always return our test file.
        self.art_file = self.temp_dir_path / "tmpcover.jpg"
        self.art_file.touch()
        self.old_afa = self.plugin.art_for_album
        self.afa_response = fetchart.Candidate(
            logger,
            source_name="test",
            path=self.art_file,
        )

        def art_for_album(i, p, local_only=False):
            return self.afa_response

        self.plugin.art_for_album = art_for_album

        # Test library.
        os.mkdir(syspath(os.path.join(self.libdir, b"album")))
        itempath = os.path.join(self.libdir, b"album", b"test.mp3")
        shutil.copyfile(
            syspath(os.path.join(_common.RSRC, b"full.mp3")),
            syspath(itempath),
        )
        self.i = _common.item()
        self.i.path = itempath
        self.album = self.lib.add_album([self.i])
        self.lib._connection().commit()

        # The import configuration.
        self.session = _common.import_session(self.lib)

        # Import task for the coroutine.
        self.task = importer.ImportTask(None, None, [self.i])
        self.task.is_album = True
        self.task.album = self.album
        info = AlbumInfo(
            album="some album",
            album_id="albumid",
            artist="some artist",
            artist_id="artistid",
            tracks=[],
        )
        self.task.set_choice(AlbumMatch(Distance(), info, {}))

    def tearDown(self):
        super().tearDown()
        self.plugin.art_for_album = self.old_afa

    def _fetch_art(self, should_exist):
        """Execute the fetch_art coroutine for the task and return the
        album's resulting artpath. ``should_exist`` specifies whether to
        assert that art path was set (to the correct value) or or that
        the path was not set.
        """
        # Execute the two relevant parts of the importer.
        self.plugin.fetch_art(self.session, self.task)
        self.plugin.assign_art(self.session, self.task)

        artpath = self.lib.albums()[0].art_filepath
        if should_exist:
            assert artpath == self.i.filepath.parent / "cover.jpg"
            assert artpath.exists()
        else:
            assert artpath is None
        return artpath

    def test_fetch_art(self):
        assert not self.lib.albums()[0].artpath
        self._fetch_art(True)

    def test_art_not_found(self):
        self.afa_response = None
        self._fetch_art(False)

    def test_no_art_for_singleton(self):
        self.task.is_album = False
        self._fetch_art(False)

    def test_leave_original_file_in_place(self):
        self._fetch_art(True)
        assert self.art_file.exists()

    def test_delete_original_file(self):
        prev_move = config["import"]["move"].get()
        try:
            config["import"]["move"] = True
            self._fetch_art(True)
            assert not self.art_file.exists()
        finally:
            config["import"]["move"] = prev_move

    def test_do_not_delete_original_if_already_in_place(self):
        artdest = os.path.join(os.path.dirname(self.i.path), b"cover.jpg")
        shutil.copyfile(self.art_file, syspath(artdest))
        self.afa_response = fetchart.Candidate(
            logger,
            source_name="test",
            path=artdest,
        )
        self._fetch_art(True)

    def test_fetch_art_if_imported_file_deleted(self):
        # See #1126. Test the following scenario:
        #   - Album art imported, `album.artpath` set.
        #   - Imported album art file subsequently deleted (by user or other
        #     program).
        # `fetchart` should import album art again instead of printing the
        # message "<album> has album art".
        self._fetch_art(True)
        util.remove(self.album.artpath)
        self.plugin.batch_fetch_art(
            self.lib, self.lib.albums(), force=False, quiet=False
        )
        assert self.album.art_filepath.exists()


class AlbumArtOperationTestCase(UseThePlugin):
    """Base test case for album art operations.

    Provides common setup for testing album art processing operations by setting
    up a mock filesystem source that returns a predefined test image.
    """

    IMAGE_PATH = os.path.join(_common.RSRC, b"abbey-similar.jpg")
    IMAGE_FILESIZE = os.stat(util.syspath(IMAGE_PATH)).st_size
    IMAGE_WIDTH = 500
    IMAGE_HEIGHT = 490
    IMAGE_WIDTH_HEIGHT_DIFF = IMAGE_WIDTH - IMAGE_HEIGHT

    @classmethod
    def setUpClass(cls):
        super().setUpClass()

        def fs_source_get(_self, album, settings, paths):
            if paths:
                yield fetchart.Candidate(
                    logger, source_name=_self.ID, path=cls.IMAGE_PATH
                )

        patch("beetsplug.fetchart.FileSystem.get", fs_source_get).start()
        cls.addClassCleanup(patch.stopall)

    def get_album_art(self):
        return self.plugin.art_for_album(_common.Bag(), [""], True)


class AlbumArtOperationConfigurationTest(AlbumArtOperationTestCase):
    """Check that scale & filesize configuration is respected.

    Depending on `minwidth`, `enforce_ratio`, `margin_px`, and `margin_percent`
    configuration the plugin should or should not return an art candidate.
    """

    def test_minwidth(self):
        self.plugin.minwidth = self.IMAGE_WIDTH / 2
        assert self.get_album_art()

        self.plugin.minwidth = self.IMAGE_WIDTH * 2
        assert not self.get_album_art()

    def test_enforce_ratio(self):
        self.plugin.enforce_ratio = True
        assert not self.get_album_art()

        self.plugin.enforce_ratio = False
        assert self.get_album_art()

    def test_enforce_ratio_with_px_margin(self):
        self.plugin.enforce_ratio = True

        self.plugin.margin_px = self.IMAGE_WIDTH_HEIGHT_DIFF * 0.5
        assert not self.get_album_art()

        self.plugin.margin_px = self.IMAGE_WIDTH_HEIGHT_DIFF * 1.5
        assert self.get_album_art()

    def test_enforce_ratio_with_percent_margin(self):
        self.plugin.enforce_ratio = True
        diff_by_width = self.IMAGE_WIDTH_HEIGHT_DIFF / self.IMAGE_WIDTH

        self.plugin.margin_percent = diff_by_width * 0.5
        assert not self.get_album_art()

        self.plugin.margin_percent = diff_by_width * 1.5
        assert self.get_album_art()


class AlbumArtPerformOperationTest(AlbumArtOperationTestCase):
    """Test that the art is resized and deinterlaced if necessary."""

    def setUp(self):
        super().setUp()
        self.resizer_mock = patch.object(
            ArtResizer.shared, "resize", return_value=self.IMAGE_PATH
        ).start()
        self.deinterlacer_mock = patch.object(
            ArtResizer.shared, "deinterlace", return_value=self.IMAGE_PATH
        ).start()

    def test_resize(self):
        self.plugin.maxwidth = self.IMAGE_WIDTH / 2
        assert self.get_album_art()
        assert self.resizer_mock.called

    def test_file_resized(self):
        self.plugin.max_filesize = self.IMAGE_FILESIZE // 2
        assert self.get_album_art()
        assert self.resizer_mock.called

    def test_file_not_resized(self):
        self.plugin.max_filesize = self.IMAGE_FILESIZE
        assert self.get_album_art()
        assert not self.resizer_mock.called

    def test_file_resized_but_not_scaled(self):
        self.plugin.maxwidth = self.IMAGE_WIDTH * 2
        self.plugin.max_filesize = self.IMAGE_FILESIZE // 2
        assert self.get_album_art()
        assert self.resizer_mock.called

    def test_file_resized_and_scaled(self):
        self.plugin.maxwidth = self.IMAGE_WIDTH / 2
        self.plugin.max_filesize = self.IMAGE_FILESIZE // 2
        assert self.get_album_art()
        assert self.resizer_mock.called

    def test_deinterlaced(self):
        self.plugin.deinterlace = True
        assert self.get_album_art()
        assert self.deinterlacer_mock.called

    def test_not_deinterlaced(self):
        self.plugin.deinterlace = False
        assert self.get_album_art()
        assert not self.deinterlacer_mock.called

    def test_deinterlaced_and_resized(self):
        self.plugin.maxwidth = self.IMAGE_WIDTH / 2
        self.plugin.deinterlace = True
        assert self.get_album_art()
        assert self.deinterlacer_mock.called
        assert self.resizer_mock.called


class DeprecatedConfigTest(unittest.TestCase):
    """While refactoring the plugin, the remote_priority option was deprecated,
    and a new codepath should translate its effect. Check that it actually does
    so.
    """

    # If we subclassed UseThePlugin, the configuration change would either be
    # overwritten by BeetsTestCase or be set after constructing the
    # plugin object
    def setUp(self):
        super().setUp()
        config["fetchart"]["remote_priority"] = True
        self.plugin = fetchart.FetchArtPlugin()

    def test_moves_filesystem_to_end(self):
        assert isinstance(self.plugin.sources[-1], fetchart.FileSystem)


class EnforceRatioConfigTest(unittest.TestCase):
    """Throw some data at the regexes."""

    def _load_with_config(self, values, should_raise):
        if should_raise:
            for v in values:
                config["fetchart"]["enforce_ratio"] = v
                with pytest.raises(confuse.ConfigValueError):
                    fetchart.FetchArtPlugin()
        else:
            for v in values:
                config["fetchart"]["enforce_ratio"] = v
                fetchart.FetchArtPlugin()

    def test_px(self):
        self._load_with_config("0px 4px 12px 123px".split(), False)
        self._load_with_config("00px stuff5px".split(), True)

    def test_percent(self):
        self._load_with_config("0% 0.00% 5.1% 5% 100%".split(), False)
        self._load_with_config("00% 1.234% foo5% 100.1%".split(), True)



# This file is part of beets.
# Copyright 2016, Adrian Sampson.
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
from __future__ import annotations

import os
import time
from typing import TYPE_CHECKING

from beets import config, logging, plugins, util
from beets.importer.tasks import Action
from beets.util import displayable_path, normpath, pipeline, syspath

from . import stages as stagefuncs
from .state import ImportState

if TYPE_CHECKING:
    from collections.abc import Sequence

    from beets import dbcore, library
    from beets.util import PathBytes

    from .tasks import ImportTask


QUEUE_SIZE = 128

# Global logger.
log = logging.getLogger("beets")


class ImportAbortError(Exception):
    """Raised when the user aborts the tagging operation."""

    pass


class ImportSession:
    """Controls an import action. Subclasses should implement methods to
    communicate with the user or otherwise make decisions.
    """

    logger: logging.Logger
    paths: list[PathBytes]
    lib: library.Library

    _is_resuming: dict[bytes, bool]
    _merged_items: set[PathBytes]
    _merged_dirs: set[PathBytes]

    def __init__(
        self,
        lib: library.Library,
        loghandler: logging.Handler | None,
        paths: Sequence[PathBytes] | None,
        query: dbcore.Query | None,
    ):
        """Create a session.

        Parameters
        ----------
        lib : library.Library
            The library instance to which items will be imported.
        loghandler : logging.Handler or None
            A logging handler to use for the session's logger. If None, a
            NullHandler will be used.
        paths : os.PathLike or None
            The paths to be imported.
        query : dbcore.Query or None
            A query to filter items for import.
        """
        self.lib = lib
        self.logger = self._setup_logging(loghandler)
        self.query = query
        self._is_resuming = {}
        self._merged_items = set()
        self._merged_dirs = set()

        # Normalize the paths.
        self.paths = list(map(normpath, paths or []))

    def _setup_logging(self, loghandler: logging.Handler | None):
        logger = logging.getLogger(__name__)
        logger.propagate = False
        if not loghandler:
            loghandler = logging.NullHandler()
        logger.handlers = [loghandler]
        return logger

    def set_config(self, config):
        """Set `config` property from global import config and make
        implied changes.
        """
        # FIXME: Maybe this function should not exist and should instead
        # provide "decision wrappers" like "should_resume()", etc.
        iconfig = dict(config)
        self.config = iconfig

        # Incremental and progress are mutually exclusive.
        if iconfig["incremental"]:
            iconfig["resume"] = False

        # When based on a query instead of directories, never
        # save progress or try to resume.
        if self.query is not None:
            iconfig["resume"] = False
            iconfig["incremental"] = False

        if iconfig["reflink"]:
            iconfig["reflink"] = iconfig["reflink"].as_choice(
                ["auto", True, False]
            )

        # Copy, move, reflink, link, and hardlink are mutually exclusive.
        if iconfig["move"]:
            iconfig["copy"] = False
            iconfig["link"] = False
            iconfig["hardlink"] = False
            iconfig["reflink"] = False
        elif iconfig["link"]:
            iconfig["copy"] = False
            iconfig["move"] = False
            iconfig["hardlink"] = False
            iconfig["reflink"] = False
        elif iconfig["hardlink"]:
            iconfig["copy"] = False
            iconfig["move"] = False
            iconfig["link"] = False
            iconfig["reflink"] = False
        elif iconfig["reflink"]:
            iconfig["copy"] = False
            iconfig["move"] = False
            iconfig["link"] = False
            iconfig["hardlink"] = False

        # Only delete when copying.
        if not iconfig["copy"]:
            iconfig["delete"] = False

        self.want_resume = config["resume"].as_choice([True, False, "ask"])

    def tag_log(self, status, paths: Sequence[PathBytes]):
        """Log a message about a given album to the importer log. The status
        should reflect the reason the album couldn't be tagged.
        """
        self.logger.info("{} {}", status, displayable_path(paths))

    def log_choice(self, task: ImportTask, duplicate=False):
        """Logs the task's current choice if it should be logged. If
        ``duplicate``, then this is a secondary choice after a duplicate was
        detected and a decision was made.
        """
        paths = task.paths
        if duplicate:
            # Duplicate: log all three choices (skip, keep both, and trump).
            if task.should_remove_duplicates:
                self.tag_log("duplicate-replace", paths)
            elif task.choice_flag in (Action.ASIS, Action.APPLY):
                self.tag_log("duplicate-keep", paths)
            elif task.choice_flag is Action.SKIP:
                self.tag_log("duplicate-skip", paths)
        else:
            # Non-duplicate: log "skip" and "asis" choices.
            if task.choice_flag is Action.ASIS:
                self.tag_log("asis", paths)
            elif task.choice_flag is Action.SKIP:
                self.tag_log("skip", paths)

    def should_resume(self, path: PathBytes):
        raise NotImplementedError

    def choose_match(self, task: ImportTask):
        raise NotImplementedError

    def resolve_duplicate(self, task: ImportTask, found_duplicates):
        raise NotImplementedError

    def choose_item(self, task: ImportTask):
        raise NotImplementedError

    def run(self):
        """Run the import task."""
        self.logger.info("import started {}", time.asctime())
        self.set_config(config["import"])

        # Set up the pipeline.
        if self.query is None:
            stages = [stagefuncs.read_tasks(self)]
        else:
            stages = [stagefuncs.query_tasks(self)]

        # In pretend mode, just log what would otherwise be imported.
        if self.config["pretend"]:
            stages += [stagefuncs.log_files(self)]
        else:
            if self.config["group_albums"] and not self.config["singletons"]:
                # Split directory tasks into one task for each album.
                stages += [stagefuncs.group_albums(self)]

            # These stages either talk to the user to get a decision or,
            # in the case of a non-autotagged import, just choose to
            # import everything as-is. In *both* cases, these stages
            # also add the music to the library database, so later
            # stages need to read and write data from there.
            if self.config["autotag"]:
                stages += [
                    stagefuncs.lookup_candidates(self),
                    stagefuncs.user_query(self),
                ]
            else:
                stages += [stagefuncs.import_asis(self)]

            # Plugin stages.
            for stage_func in plugins.early_import_stages():
                stages.append(stagefuncs.plugin_stage(self, stage_func))
            for stage_func in plugins.import_stages():
                stages.append(stagefuncs.plugin_stage(self, stage_func))

            stages += [stagefuncs.manipulate_files(self)]

        pl = pipeline.Pipeline(stages)

        # Run the pipeline.
        plugins.send("import_begin", session=self)
        try:
            if config["threaded"]:
                pl.run_parallel(QUEUE_SIZE)
            else:
                pl.run_sequential()
        except ImportAbortError:
            # User aborted operation. Silently stop.
            pass

    # Incremental and resumed imports

    def already_imported(self, toppath: PathBytes, paths: Sequence[PathBytes]):
        """Returns true if the files belonging to this task have already
        been imported in a previous session.
        """
        if self.is_resuming(toppath) and all(
            [ImportState().progress_has_element(toppath, p) for p in paths]
        ):
            return True
        if self.config["incremental"] and tuple(paths) in self.history_dirs:
            return True

        return False

    _history_dirs = None

    @property
    def history_dirs(self) -> set[tuple[PathBytes, ...]]:
        # FIXME: This could be simplified to a cached property
        if self._history_dirs is None:
            self._history_dirs = ImportState().taghistory
        return self._history_dirs

    def already_merged(self, paths: Sequence[PathBytes]):
        """Returns true if all the paths being imported were part of a merge
        during previous tasks.
        """
        for path in paths:
            if path not in self._merged_items and path not in self._merged_dirs:
                return False
        return True

    def mark_merged(self, paths: Sequence[PathBytes]):
        """Mark paths and directories as merged for future reimport tasks."""
        self._merged_items.update(paths)
        dirs = {
            os.path.dirname(path) if os.path.isfile(syspath(path)) else path
            for path in paths
        }
        self._merged_dirs.update(dirs)

    def is_resuming(self, toppath: PathBytes):
        """Return `True` if user wants to resume import of this path.

        You have to call `ask_resume` first to determine the return value.
        """
        return self._is_resuming.get(toppath, False)

    def ask_resume(self, toppath: PathBytes):
        """If import of `toppath` was aborted in an earlier session, ask
        user if they want to resume the import.

        Determines the return value of `is_resuming(toppath)`.
        """
        if self.want_resume and ImportState().progress_has(toppath):
            # Either accept immediately or prompt for input to decide.
            if self.want_resume is True or self.should_resume(toppath):
                log.warning(
                    "Resuming interrupted import of {}",
                    util.displayable_path(toppath),
                )
                self._is_resuming[toppath] = True
            else:
                # Clear progress; we're starting from the top.
                ImportState().progress_reset(toppath)



# This file is part of beets.
# Copyright 2016, Heinz Wiesinger.
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.

"""Synchronize information from amarok's library via dbus"""

from datetime import datetime
from os.path import basename
from time import mktime
from typing import ClassVar
from xml.sax.saxutils import quoteattr

from beets.dbcore import types
from beets.util import displayable_path
from beetsplug.metasync import MetaSource


def import_dbus():
    try:
        return __import__("dbus")
    except ImportError:
        return None


dbus = import_dbus()


class Amarok(MetaSource):
    item_types: ClassVar[dict[str, types.Type]] = {
        "amarok_rating": types.INTEGER,
        "amarok_score": types.FLOAT,
        "amarok_uid": types.STRING,
        "amarok_playcount": types.INTEGER,
        "amarok_firstplayed": types.DATE,
        "amarok_lastplayed": types.DATE,
    }

    query_xml = """
        <query version="1.0">
            <filters>
                <and><include field="filename" value={} /></and>
            </filters>
        </query>"""

    def __init__(self, config, log):
        super().__init__(config, log)

        if not dbus:
            raise ImportError("failed to import dbus")

        self.collection = dbus.SessionBus().get_object(
            "org.kde.amarok", "/Collection"
        )

    def sync_from_source(self, item):
        path = displayable_path(item.path)

        # amarok unfortunately doesn't allow searching for the full path, only
        # for the patch relative to the mount point. But the full path is part
        # of the result set. So query for the filename and then try to match
        # the correct item from the results we get back
        results = self.collection.Query(
            self.query_xml.format(quoteattr(basename(path)))
        )
        for result in results:
            if result["xesam:url"] != path:
                continue

            item.amarok_rating = result["xesam:userRating"]
            item.amarok_score = result["xesam:autoRating"]
            item.amarok_playcount = result["xesam:useCount"]
            item.amarok_uid = result["xesam:id"].replace(
                "amarok-sqltrackuid://", ""
            )

            if result["xesam:firstUsed"][0][0] != 0:
                # These dates are stored as timestamps in amarok's db, but
                # exposed over dbus as fixed integers in the current timezone.
                first_played = datetime(
                    result["xesam:firstUsed"][0][0],
                    result["xesam:firstUsed"][0][1],
                    result["xesam:firstUsed"][0][2],
                    result["xesam:firstUsed"][1][0],
                    result["xesam:firstUsed"][1][1],
                    result["xesam:firstUsed"][1][2],
                )

                if result["xesam:lastUsed"][0][0] != 0:
                    last_played = datetime(
                        result["xesam:lastUsed"][0][0],
                        result["xesam:lastUsed"][0][1],
                        result["xesam:lastUsed"][0][2],
                        result["xesam:lastUsed"][1][0],
                        result["xesam:lastUsed"][1][1],
                        result["xesam:lastUsed"][1][2],
                    )
                else:
                    last_played = first_played

                item.amarok_firstplayed = mktime(first_played.timetuple())
                item.amarok_lastplayed = mktime(last_played.timetuple())



# This file is part of beets.
# Copyright 2015, winters jean-marie.
# Copyright 2020, Justin Mayer <https://justinmayer.com>
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.

"""This plugin generates tab completions for Beets commands for the Fish shell
<https://fishshell.com/>, including completions for Beets commands, plugin
commands, and option flags. Also generated are completions for all the album
and track fields, suggesting for example `genres:` or `album:` when querying the
Beets database. Completions for the *values* of those fields are not generated
by default but can be added via the `-e` / `--extravalues` flag. For example:
`beet fish -e genres -e albumartist`
"""

import os
from operator import attrgetter

from beets import library, plugins, ui
from beets.plugins import BeetsPlugin
from beets.ui import commands

BL_NEED2 = """complete -c beet -n '__fish_beet_needs_command' {} {}\n"""
BL_USE3 = """complete -c beet -n '__fish_beet_using_command {}' {} {}\n"""
BL_SUBS = """complete -c beet -n '__fish_at_level {} ""' {}  {}\n"""
BL_EXTRA3 = """complete -c beet -n '__fish_beet_use_extra {}' {} {}\n"""

HEAD = """
function __fish_beet_needs_command
    set cmd (commandline -opc)
    if test (count $cmd) -eq 1
        return 0
    end
    return 1
end

function __fish_beet_using_command
    set cmd (commandline -opc)
    set needle (count $cmd)
    if test $needle -gt 1
        if begin test $argv[1] = $cmd[2];
            and not contains -- $cmd[$needle] $FIELDS; end
                return 0
        end
    end
    return 1
end

function __fish_beet_use_extra
    set cmd (commandline -opc)
    set needle (count $cmd)
    if test $argv[2]  = $cmd[$needle]
        return 0
    end
    return 1
end
"""


class FishPlugin(BeetsPlugin):
    def commands(self):
        cmd = ui.Subcommand("fish", help="generate Fish shell tab completions")
        cmd.func = self.run
        cmd.parser.add_option(
            "-f",
            "--noFields",
            action="store_true",
            default=False,
            help="omit album/track field completions",
        )
        cmd.parser.add_option(
            "-e",
            "--extravalues",
            action="append",
            type="choice",
            choices=library.Item.all_keys() + library.Album.all_keys(),
            help="include specified field *values* in completions",
        )
        cmd.parser.add_option(
            "-o",
            "--output",
            default="~/.config/fish/completions/beet.fish",
            help=(
                "where to save the script. default: ~/.config/fish/completions"
            ),
        )
        return [cmd]

    def run(self, lib, opts, args):
        # Gather the commands from Beets core and its plugins.
        # Collect the album and track fields.
        # If specified, also collect the values for these fields.
        # Make a giant string of all the above, formatted in a way that
        # allows Fish to do tab completion for the `beet` command.

        completion_file_path = os.path.expanduser(opts.output)
        completion_dir = os.path.dirname(completion_file_path)

        if completion_dir != "":
            os.makedirs(completion_dir, exist_ok=True)

        nobasicfields = opts.noFields  # Do not complete for album/track fields
        extravalues = opts.extravalues  # e.g., Also complete artists names
        beetcmds = sorted(
            (commands.default_commands + plugins.commands()),
            key=attrgetter("name"),
        )
        fields = sorted(set(library.Album.all_keys() + library.Item.all_keys()))
        # Collect commands, their aliases, and their help text
        cmd_names_help = []
        for cmd in beetcmds:
            names = list(cmd.aliases)
            names.append(cmd.name)
            for name in names:
                cmd_names_help.append((name, cmd.help))
        # Concatenate the string
        totstring = f"{HEAD}\n"
        totstring += get_cmds_list([name[0] for name in cmd_names_help])
        totstring += "" if nobasicfields else get_standard_fields(fields)
        totstring += get_extravalues(lib, extravalues) if extravalues else ""
        totstring += "\n# ====== setup basic beet completion =====\n\n"
        totstring += get_basic_beet_options()
        totstring += "\n# ====== setup field completion for subcommands =====\n"
        totstring += get_subcommands(cmd_names_help, nobasicfields, extravalues)
        # Set up completion for all the command options
        totstring += get_all_commands(beetcmds)

        with open(completion_file_path, "w") as fish_file:
            fish_file.write(totstring)


def _escape(name):
    # Escape ? in fish
    if name == "?":
        name = f"\\{name}"
    return name


def get_cmds_list(cmds_names):
    # Make a list of all Beets core & plugin commands
    return f"set CMDS {' '.join(cmds_names)}\n\n"


def get_standard_fields(fields):
    # Make a list of album/track fields and append with ':'
    fields = (f"{field}:" for field in fields)
    return f"set FIELDS {' '.join(fields)}\n\n"


def get_extravalues(lib, extravalues):
    # Make a list of all values from an album/track field.
    # 'beet ls albumartist: <TAB>' yields completions for ABBA, Beatles, etc.
    word = ""
    values_set = get_set_of_values_for_field(lib, extravalues)
    for fld in extravalues:
        extraname = f"{fld.upper()}S"
        word += f"set  {extraname} {' '.join(sorted(values_set[fld]))}\n\n"
    return word


def get_set_of_values_for_field(lib, fields):
    # Get unique values from a specified album/track field
    fields_dict = {}
    for each in fields:
        fields_dict[each] = set()
    for item in lib.items():
        for field in fields:
            fields_dict[field].add(wrap(item[field]))
    return fields_dict


def get_basic_beet_options():
    word = (
        BL_NEED2.format("-l format-item", "-f -d 'print with custom format'")
        + BL_NEED2.format("-l format-album", "-f -d 'print with custom format'")
        + BL_NEED2.format(
            "-s  l  -l library", "-F -r -d 'library database file to use'"
        )
        + BL_NEED2.format(
            "-s  d  -l directory", "-F -r -d 'destination music directory'"
        )
        + BL_NEED2.format(
            "-s  v  -l verbose", "-f -d 'print debugging information'"
        )
        + BL_NEED2.format(
            "-s  c  -l config", "-F -r -d 'path to configuration file'"
        )
        + BL_NEED2.format(
            "-s  h  -l help", "-f -d 'print this help message and exit'"
        )
    )
    return word


def get_subcommands(cmd_name_and_help, nobasicfields, extravalues):
    # Formatting for Fish to complete our fields/values
    word = ""
    for cmdname, cmdhelp in cmd_name_and_help:
        cmdname = _escape(cmdname)

        word += f"\n# ------ fieldsetups for {cmdname} -------\n"
        word += BL_NEED2.format(
            f"-a {cmdname}", f"-f -d {wrap(clean_whitespace(cmdhelp))}"
        )

        if nobasicfields is False:
            word += BL_USE3.format(
                cmdname,
                f"-a {wrap('$FIELDS')}",
                f"-d {wrap('fieldname')}",
            )

        if extravalues:
            for f in extravalues:
                setvar = wrap(f"${f.upper()}S")
                word += " ".join(
                    BL_EXTRA3.format(
                        f"{cmdname} {f}:",
                        f"-f -A -a {setvar}",
                        f"-d {wrap(f)}",
                    ).split()
                )
                word += "\n"
    return word


def get_all_commands(beetcmds):
    # Formatting for Fish to complete command options
    word = ""
    for cmd in beetcmds:
        names = list(cmd.aliases)
        names.append(cmd.name)
        for name in names:
            name = _escape(name)

            word += f"\n\n\n# ====== completions for {name} =====\n"

            for option in cmd.parser._get_all_options()[1:]:
                cmd_l = (
                    f" -l {option._long_opts[0].replace('--', '')}"
                    if option._long_opts
                    else ""
                )
                cmd_s = (
                    f" -s {option._short_opts[0].replace('-', '')}"
                    if option._short_opts
                    else ""
                )
                cmd_need_arg = " -r " if option.nargs in [1] else ""
                cmd_helpstr = (
                    f" -d {wrap(' '.join(option.help.split()))}"
                    if option.help
                    else ""
                )
                cmd_arglist = (
                    f" -a {wrap(' '.join(option.choices))}"
                    if option.choices
                    else ""
                )

                word += " ".join(
                    BL_USE3.format(
                        name,
                        f"{cmd_need_arg}{cmd_s}{cmd_l} {cmd_arglist}",
                        cmd_helpstr,
                    ).split()
                )
                word += "\n"

            word = word + BL_USE3.format(
                name,
                "-s h -l help",
                f"-d {wrap('print help')}",
            )
    return word


def clean_whitespace(word):
    # Remove excess whitespace and tabs in a string
    return " ".join(word.split())


def wrap(word):
    # Need " or ' around strings but watch out if they're in the string
    sptoken = '"'
    if '"' in word and ("'") in word:
        word.replace('"', sptoken)
        return f'"{word}"'

    tok = '"' if "'" in word else "'"
    return f"{tok}{word}{tok}"



# This file is part of beets.
# Copyright 2016, Adrian Sampson.
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.

"""Tests for autotagging functionality."""

import operator
from contextlib import nullcontext as does_not_warn

import pytest

from beets.autotag.distance import Distance
from beets.autotag.hooks import (
    AlbumInfo,
    AlbumMatch,
    TrackInfo,
    TrackMatch,
    correct_list_fields,
)
from beets.library import Item
from beets.test.helper import BeetsTestCase

str_field_deprecation = pytest.warns(
    DeprecationWarning, match="The 'genre' field is deprecated"
)

_p = pytest.param


@pytest.mark.parametrize(
    "str_value, list_value, expected_warning, expected_list_value",
    [
        _p(
            "value",
            None,
            str_field_deprecation,
            ["value"],
            id="str value only, warning raised",
        ),
        _p(
            "value; another value",
            None,
            str_field_deprecation,
            ["value", "another value"],
            id="str value only and split, warning raised",
        ),
        _p(
            "value",
            ["list value"],
            str_field_deprecation,
            ["list value"],
            id="list value wins, warning raised",
        ),
        _p(
            None,
            None,
            does_not_warn(),
            None,
            id="no str value, no warning",
        ),
    ],
)
class TestLegacyStringField:
    def test_init_info(
        self, str_value, list_value, expected_warning, expected_list_value
    ):
        with expected_warning:
            actual_list_value = TrackInfo._get_list_from_string_value(
                "genre", "genres", str_value, list_value
            )

        assert actual_list_value == expected_list_value

    def test_set_str_value(
        self, str_value, list_value, expected_warning, expected_list_value
    ):
        info = TrackInfo(genres=list_value)
        with expected_warning:
            info["genre"] = str_value

        assert info["genres"] == expected_list_value


class ApplyTest(BeetsTestCase):
    def _apply(
        self,
        per_disc_numbering=False,
        artist_credit=False,
        original_date=False,
        from_scratch=False,
    ):
        info = self.info
        mapping = dict(zip(self.items, info.tracks))
        self.config["per_disc_numbering"] = per_disc_numbering
        self.config["artist_credit"] = artist_credit
        self.config["original_date"] = original_date
        self.config["import"]["from_scratch"] = from_scratch
        amatch = AlbumMatch(Distance(), self.info, mapping)
        amatch.apply_metadata()
        return amatch

    def setUp(self):
        super().setUp()

        self.items = [Item(), Item()]
        self.info = AlbumInfo(
            tracks=[
                TrackInfo(
                    title="title",
                    track_id="dfa939ec-118c-4d0f-84a0-60f3d1e6522c",
                    medium=1,
                    medium_index=1,
                    medium_total=1,
                    index=1,
                    artist="trackArtist",
                    artist_credit="trackArtistCredit",
                    artists_credit=["trackArtistCredit"],
                    artist_sort="sortLast, sortFirst and sortLast2, sortFirst2",
                    artists_sort=[
                        "sortLast, sortFirst",
                        "sortLast2, sortFirst2",
                    ],
                    arranger="trackArranger1, trackArranger2",
                    composer="trackComposer1, trackComposer2",
                    genres=["Rock"],
                    lyricist="trackLyricist1, trackLyricist2",
                    remixer="trackRemixer1, trackRemixer2",
                ),
                TrackInfo(
                    title="title2",
                    track_id="40130ed1-a27c-42fd-a328-1ebefb6caef4",
                    medium=2,
                    medium_index=1,
                    index=2,
                    medium_total=1,
                    arrangers=["trackArranger1", "trackArranger2"],
                    composers=["trackComposer1", "trackComposer2"],
                    lyricists=["trackLyricist1", "trackLyricist2"],
                    remixers=["trackRemixer1", "trackRemixer2"],
                ),
            ],
            artist="albumArtist",
            artists=["albumArtist", "albumArtist2"],
            album="album",
            album_id="7edb51cb-77d6-4416-a23c-3a8c2994a2c7",
            artist_id="a6623d39-2d8e-4f70-8242-0a9553b91e50",
            artists_ids=None,
            artist_credit="albumArtistCredit",
            artists_credit=["albumArtistCredit1", "albumArtistCredit2"],
            artist_sort=None,
            artists_sort=["albumArtistSort", "albumArtistSort2"],
            albumtype="album",
            va=True,
            mediums=2,
            data_source="MusicBrainz",
            year=2013,
            month=12,
            day=18,
            genre="Rock; Pop",
        )

        common_expected = {
            "album": "album",
            "albumartist_credit": "albumArtistCredit",
            "albumartist": "albumArtist",
            "albumartists": ["albumArtist", "albumArtist2"],
            "albumartists_credit": [
                "albumArtistCredit",
                "albumArtistCredit1",
                "albumArtistCredit2",
            ],
            "albumartist_sort": "albumArtistSort",
            "albumartists_sort": ["albumArtistSort", "albumArtistSort2"],
            "albumtype": "album",
            "albumtypes": ["album"],
            "comp": True,
            "disctotal": 2,
            "mb_albumartistid": "a6623d39-2d8e-4f70-8242-0a9553b91e50",
            "mb_albumartistids": ["a6623d39-2d8e-4f70-8242-0a9553b91e50"],
            "mb_albumid": "7edb51cb-77d6-4416-a23c-3a8c2994a2c7",
            "mb_artistid": "a6623d39-2d8e-4f70-8242-0a9553b91e50",
            "mb_artistids": ["a6623d39-2d8e-4f70-8242-0a9553b91e50"],
            "arrangers": ["trackArranger1", "trackArranger2"],
            "composers": ["trackComposer1", "trackComposer2"],
            "remixers": ["trackRemixer1", "trackRemixer2"],
            "lyricists": ["trackLyricist1", "trackLyricist2"],
            "tracktotal": 2,
            "year": 2013,
            "month": 12,
            "day": 18,
        }

        self.expected_tracks = [
            {
                **common_expected,
                "artist": "trackArtist",
                "artists": ["trackArtist"],
                "artist_credit": "trackArtistCredit",
                "artist_sort": "sortLast, sortFirst and sortLast2, sortFirst2",
                "artists_credit": ["trackArtistCredit"],
                "artists_sort": [
                    "sortLast, sortFirst",
                    "sortLast2, sortFirst2",
                ],
                "disc": 1,
                "genres": ["Rock"],
                "mb_trackid": "dfa939ec-118c-4d0f-84a0-60f3d1e6522c",
                "title": "title",
                "track": 1,
            },
            {
                **common_expected,
                "artist": "albumArtist",
                "artists": ["albumArtist", "albumArtist2"],
                "artist_credit": "albumArtistCredit",
                "artist_sort": "albumArtistSort",
                "artists_credit": [
                    "albumArtistCredit",
                    "albumArtistCredit1",
                    "albumArtistCredit2",
                ],
                "artists_sort": ["albumArtistSort", "albumArtistSort2"],
                "disc": 2,
                "genres": ["Rock", "Pop"],
                "mb_trackid": "40130ed1-a27c-42fd-a328-1ebefb6caef4",
                "title": "title2",
                "track": 2,
            },
        ]

    def test_autotag_items(self):
        self._apply()

        keys = self.expected_tracks[0].keys()
        get_values = operator.itemgetter(*keys)

        applied_data = [
            dict(zip(keys, get_values(dict(i)))) for i in self.items
        ]

        assert applied_data == self.expected_tracks

    def test_per_disc_numbering(self):
        self._apply(per_disc_numbering=True)

        assert self.items[0].track == 1
        assert self.items[1].track == 1
        assert self.items[0].tracktotal == 1
        assert self.items[1].tracktotal == 1

    def test_artist_credit_prefers_artist_over_albumartist_credit(self):
        self.info.tracks[0].update(artist="oldArtist", artist_credit=None)

        self._apply(artist_credit=True)

        assert self.items[0].artist == "oldArtist"

    def test_artist_credit_falls_back_to_albumartist(self):
        self.info.artist_credit = None

        self._apply(artist_credit=True)

        assert self.items[1].artist == "albumArtist"

    def test_date_only_zeroes_month_and_day(self):
        self.items = [Item(year=1, month=2, day=3)]
        self.info.update(year=2013, month=None, day=None)

        self._apply()

        assert self.items[0].year == 2013
        assert self.items[0].month == 0
        assert self.items[0].day == 0

    def test_missing_date_applies_nothing(self):
        self.items = [Item(year=1, month=2, day=3)]
        self.info.update(year=None, month=None, day=None)

        self._apply()

        assert self.items[0].year == 1
        assert self.items[0].month == 2
        assert self.items[0].day == 3

    def test_original_date_overrides_album_metadata(self):
        self.info.update(
            year=2013,
            month=12,
            day=18,
            original_year=1999,
            original_month=4,
            original_day=7,
        )
        match = self._apply(original_date=True)
        album = self.lib.add_album(self.items)
        match.apply_album_metadata(album)

        assert album.year == 1999
        assert album.month == 4
        assert album.day == 7


class TestFromScratch:
    @pytest.fixture(autouse=True)
    def config(self, config):
        config["import"]["from_scratch"] = True

    @pytest.fixture
    def album_info(self):
        return AlbumInfo(
            tracks=[TrackInfo(title="title", artist="track artist", index=1)]
        )

    @pytest.fixture
    def item(self):
        return Item(artist="old artist", comments="stale comment")

    def test_album_match_clears_stale_metadata(self, album_info, item):
        match = AlbumMatch(Distance(), album_info, {item: album_info.tracks[0]})

        match.apply_metadata()

        assert item.artist == "track artist"
        assert item.comments == ""

    def test_singleton_match_clears_stale_metadata(self, item):
        match = TrackMatch(Distance(), TrackInfo(artist="track artist"), item)

        match.apply_metadata()

        assert item.artist == "track artist"
        assert item.comments == ""


@pytest.mark.parametrize(
    "overwrite_fields,expected_item_artist",
    [
        pytest.param(["artist"], "", id="overwrite artist"),
        pytest.param([], "artist", id="do not overwrite artist"),
    ],
)
class TestOverwriteNull:
    @pytest.fixture(autouse=True)
    def config(self, config, overwrite_fields):
        config["overwrite_null"]["album"] = overwrite_fields
        config["overwrite_null"]["track"] = overwrite_fields
        config["import"]["from_scratch"] = False

    @pytest.fixture
    def item(self):
        return Item(artist="artist")

    @pytest.fixture
    def track_info(self):
        return TrackInfo(artist=None)

    def test_album(self, item, track_info, expected_item_artist):
        match = AlbumMatch(
            Distance(), AlbumInfo([track_info]), {item: track_info}
        )

        match.apply_metadata()

        assert item.artist == expected_item_artist

    def test_singleton(self, item, track_info, expected_item_artist):
        match = TrackMatch(Distance(), track_info, item)

        match.apply_metadata()

        assert item.artist == expected_item_artist


@pytest.mark.parametrize(
    "single_field,list_field",
    [
        ("albumtype", "albumtypes"),
        ("artist", "artists"),
        ("artist_credit", "artists_credit"),
        ("artist_id", "artists_ids"),
        ("artist_sort", "artists_sort"),
    ],
)
@pytest.mark.parametrize(
    "single_value,list_value,expected_values",
    [
        (None, [], (None, [])),
        (None, ["1"], ("1", ["1"])),
        (None, ["1", "2"], ("1", ["1", "2"])),
        ("1", [], ("1", ["1"])),
        ("1", ["1"], ("1", ["1"])),
        ("1", ["1", "2"], ("1", ["1", "2"])),
        ("1", ["2", "1"], ("1", ["1", "2"])),
        ("1", ["2"], ("1", ["1", "2"])),
        ("1 ft 2", ["1", "1 ft 2"], ("1 ft 2", ["1 ft 2", "1"])),
        ("1 FT 2", ["1", "1 ft 2"], ("1 FT 2", ["1", "1 ft 2"])),
        ("a", ["b", "A"], ("a", ["b", "A"])),
        ("1 ft 2", ["2", "1"], ("1 ft 2", ["2", "1"])),
        (
            "a, b and c, d",
            ["a, b", "c, d"],
            ("a, b and c, d", ["a, b", "c, d"]),
        ),
    ],
)
def test_correct_list_fields(
    single_field, list_field, single_value, list_value, expected_values
):
    """Verify that singular and plural field variants are kept consistent.

    Checks that when both a single-value field and its list counterpart are
    present, the function reconciles them: ensuring the single value appears
    in the list and the list drives the canonical single value when needed.
    """
    input_data = {single_field: single_value, list_field: list_value}

    data = correct_list_fields(input_data)

    assert (data[single_field], data[list_field]) == expected_values



# This file is part of beets.
# Copyright 2016, Jakob Schnitzer.
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.

"""Converts tracks or albums to external directory"""

import logging
import os
import shlex
import subprocess
import tempfile
import threading
from string import Template

import mediafile
from confuse import ConfigTypeError, Optional

from beets import config, plugins, ui, util
from beets.library import Item, parse_query_string
from beets.plugins import BeetsPlugin
from beets.util import par_map
from beets.util.artresizer import ArtResizer
from beets.util.m3u import M3UFile
from beetsplug._utils import art

_fs_lock = threading.Lock()
_temp_files = []  # Keep track of temporary transcoded files for deletion.

# Some convenient alternate names for formats.
ALIASES = {
    "windows media": "wma",
    "vorbis": "ogg",
}

LOSSLESS_FORMATS = ["ape", "flac", "alac", "wave", "aiff"]


def replace_ext(path, ext):
    """Return the path with its extension replaced by `ext`.

    The new extension must not contain a leading dot.
    """
    ext_dot = b"." + ext
    return os.path.splitext(path)[0] + ext_dot


def get_format(fmt=None):
    """Return the command template and the extension from the config."""
    if not fmt:
        fmt = config["convert"]["format"].as_str().lower()
    fmt = ALIASES.get(fmt, fmt)

    try:
        format_info = config["convert"]["formats"][fmt].get(dict)
        command = format_info["command"]
        extension = format_info.get("extension", fmt)
    except KeyError:
        raise ui.UserError(f'convert: format {fmt} needs the "command" field')
    except ConfigTypeError:
        command = config["convert"]["formats"][fmt].get(str)
        extension = fmt

    # Convenience and backwards-compatibility shortcuts.
    keys = config["convert"].keys()
    if "command" in keys:
        command = config["convert"]["command"].as_str()
    elif "opts" in keys:
        # Undocumented option for backwards compatibility with < 1.3.1.
        command = (
            f"ffmpeg -i $source -y {config['convert']['opts'].as_str()} $dest"
        )
    if "extension" in keys:
        extension = config["convert"]["extension"].as_str()

    return (command.encode("utf-8"), extension.encode("utf-8"))


def in_no_convert(item: Item) -> bool:
    no_convert_query = config["convert"]["no_convert"].as_str()

    if no_convert_query:
        query, _ = parse_query_string(no_convert_query, Item)
        return query.match(item)
    else:
        return False


def should_transcode(item, fmt, force: bool = False):
    """Determine whether the item should be transcoded as part of
    conversion (i.e., its bitrate is high or it has the wrong format).

    If ``force`` is True, safety checks like ``no_convert`` and
    ``never_convert_lossy_files`` are ignored and the item is always
    transcoded.
    """
    if force:
        return True
    if in_no_convert(item) or (
        config["convert"]["never_convert_lossy_files"].get(bool)
        and item.format.lower() not in LOSSLESS_FORMATS
    ):
        return False
    maxbr = config["convert"]["max_bitrate"].get(Optional(int))
    if maxbr is not None and item.bitrate >= 1000 * maxbr:
        return True
    return fmt.lower() != item.format.lower()


class ConvertPlugin(BeetsPlugin):
    def __init__(self):
        super().__init__()
        self.config.add(
            {
                "dest": None,
                "pretend": False,
                "link": False,
                "hardlink": False,
                "threads": os.cpu_count(),
                "format": "mp3",
                "id3v23": "inherit",
                "write_metadata": True,
                "formats": {
                    "aac": {
                        "command": (
                            "ffmpeg -i $source -y -vn -acodec aac -aq 1 $dest"
                        ),
                        "extension": "m4a",
                    },
                    "alac": {
                        "command": (
                            "ffmpeg -i $source -y -vn -acodec alac $dest"
                        ),
                        "extension": "m4a",
                    },
                    "flac": "ffmpeg -i $source -y -vn -acodec flac $dest",
                    "mp3": "ffmpeg -i $source -y -vn -aq 2 $dest",
                    "opus": (
                        "ffmpeg -i $source -y -vn -acodec libopus -ab 96k $dest"
                    ),
                    "ogg": (
                        "ffmpeg -i $source -y -vn -acodec libvorbis -aq 3 $dest"
                    ),
                    "wma": "ffmpeg -i $source -y -vn -acodec wmav2 -vn $dest",
                },
                "max_bitrate": None,
                "auto": False,
                "auto_keep": False,
                "tmpdir": None,
                "quiet": False,
                "embed": True,
                "paths": {},
                "no_convert": "",
                "never_convert_lossy_files": False,
                "copy_album_art": False,
                "album_art_maxwidth": 0,
                "delete_originals": False,
                "playlist": None,
            }
        )
        self.early_import_stages = [self.auto_convert, self.auto_convert_keep]

        self.register_listener("import_task_files", self._cleanup)

    def commands(self):
        cmd = ui.Subcommand("convert", help="convert to external location")
        cmd.parser.add_option(
            "-p",
            "--pretend",
            action="store_true",
            help="show actions but do nothing",
        )
        cmd.parser.add_option(
            "-t",
            "--threads",
            action="store",
            type="int",
            help=(
                "change the number of threads, defaults to maximum available"
                " processors"
            ),
        )
        cmd.parser.add_option(
            "-k",
            "--keep-new",
            action="store_true",
            dest="keep_new",
            help="keep only the converted and move the old files",
        )
        cmd.parser.add_option(
            "-d", "--dest", action="store", help="set the destination directory"
        )
        cmd.parser.add_option(
            "-f",
            "--format",
            action="store",
            dest="format",
            help="set the target format of the tracks",
        )
        cmd.parser.add_option(
            "-y",
            "--yes",
            action="store_true",
            dest="yes",
            help="do not ask for confirmation",
        )
        cmd.parser.add_option(
            "-l",
            "--link",
            action="store_true",
            dest="link",
            help="symlink files that do not need transcoding.",
        )
        cmd.parser.add_option(
            "-H",
            "--hardlink",
            action="store_true",
            dest="hardlink",
            help=(
                "hardlink files that do not need transcoding. Overrides --link."
            ),
        )
        cmd.parser.add_option(
            "-m",
            "--playlist",
            action="store",
            help="""create an m3u8 playlist file containing
                              the converted files. The playlist file will be
                              saved below the destination directory, thus
                              PLAYLIST could be a file name or a relative path.
                              To ensure a working playlist when transferred to
                              a different computer, or opened from an external
                              drive, relative paths pointing to media files
                              will be used.""",
        )
        cmd.parser.add_option(
            "-F",
            "--force",
            action="store_true",
            dest="force",
            help=(
                "force transcoding. Ignores no_convert, "
                "never_convert_lossy_files, and max_bitrate"
            ),
        )
        cmd.parser.add_album_option()
        cmd.func = self.convert_func
        return [cmd]

    def auto_convert(self, config, task):
        if self.config["auto"]:
            par_map(
                lambda item: self.convert_on_import(config.lib, item),
                task.imported_items(),
            )

    def auto_convert_keep(self, config, task):
        if self.config["auto_keep"]:
            empty_opts = self.commands()[0].parser.get_default_values()
            (
                dest,
                threads,
                path_formats,
                fmt,
                pretend,
                hardlink,
                link,
                _,
                force,
            ) = self._get_opts_and_config(empty_opts)

            items = task.imported_items()

            # Filter items based on should_transcode function
            items = [item for item in items if should_transcode(item, fmt)]

            self._parallel_convert(
                dest,
                False,
                path_formats,
                fmt,
                pretend,
                link,
                hardlink,
                threads,
                items,
                force,
            )

    # Utilities converted from functions to methods on logging overhaul

    def encode(self, command, source, dest, pretend=False):
        """Encode `source` to `dest` using command template `command`.

        Raises `subprocess.CalledProcessError` if the command exited with a
        non-zero status code.
        """
        # The paths and arguments must be bytes.
        assert isinstance(command, bytes)
        assert isinstance(source, bytes)
        assert isinstance(dest, bytes)

        quiet = self.config["quiet"].get(bool)

        if not quiet and not pretend:
            self._log.info("Encoding {}", util.displayable_path(source))

        command = os.fsdecode(command)
        source = os.fsdecode(source)
        dest = os.fsdecode(dest)

        # Substitute $source and $dest in the argument list.
        args = shlex.split(command)
        encode_cmd = []
        for i, arg in enumerate(args):
            args[i] = Template(arg).safe_substitute(
                {
                    "source": source,
                    "dest": dest,
                }
            )
            encode_cmd.append(os.fsdecode(args[i]))

        if pretend:
            self._log.info("{}", " ".join(args))
            return

        try:
            util.command_output(encode_cmd)
        except subprocess.CalledProcessError as exc:
            # Something went wrong (probably Ctrl+C), remove temporary files
            self._log.info(
                "Encoding {} failed. Cleaning up...",
                util.displayable_path(source),
            )
            self._log.debug(
                "Command {0} exited with status {1.returncode}: {1.output}",
                args,
                exc,
            )
            util.remove(dest)
            util.prune_dirs(os.path.dirname(dest))
            raise
        except OSError as exc:
            raise ui.UserError(
                f"convert: couldn't invoke {' '.join(args)!r}: {exc}"
            )

        if not quiet and not pretend:
            self._log.info(
                "Finished encoding {}", util.displayable_path(source)
            )

    def convert_item(
        self,
        dest_dir,
        keep_new,
        path_formats,
        fmt,
        pretend=False,
        link=False,
        hardlink=False,
        force=False,
    ):
        """A pipeline thread that converts `Item` objects from a
        library.
        """
        command, ext = get_format(fmt)
        item, original, converted = None, None, None
        while True:
            item = yield (item, original, converted)
            dest = item.destination(basedir=dest_dir, path_formats=path_formats)

            # Ensure that desired item is readable before processing it. Needed
            # to avoid any side-effect of the conversion (linking, keep_new,
            # refresh) if we already know that it will fail.
            try:
                mediafile.MediaFile(util.syspath(item.path))
            except mediafile.UnreadableFileError as exc:
                self._log.error("Could not open file to convert: {}", exc)
                continue

            # When keeping the new file in the library, we first move the
            # current (pristine) file to the destination. We'll then copy it
            # back to its old path or transcode it to a new path.
            if keep_new:
                original = dest
                converted = item.path
                if should_transcode(item, fmt, force):
                    converted = replace_ext(converted, ext)
            else:
                original = item.path
                if should_transcode(item, fmt, force):
                    dest = replace_ext(dest, ext)
                converted = dest

            # Ensure that only one thread tries to create directories at a
            # time. (The existence check is not atomic with the directory
            # creation inside this function.)
            if not pretend:
                with _fs_lock:
                    util.mkdirall(dest)

            if os.path.exists(util.syspath(dest)):
                self._log.info(
                    "Skipping {.filepath} (target file exists)", item
                )
                continue

            if keep_new:
                if pretend:
                    self._log.info(
                        "mv {.filepath} {}",
                        item,
                        util.displayable_path(original),
                    )
                else:
                    self._log.info(
                        "Moving to {}", util.displayable_path(original)
                    )
                    util.move(item.path, original)

            if should_transcode(item, fmt, force):
                linked = False
                try:
                    self.encode(command, original, converted, pretend)
                except subprocess.CalledProcessError:
                    continue
            else:
                linked = link or hardlink
                if pretend:
                    msg = "ln" if hardlink else ("ln -s" if link else "cp")

                    self._log.info(
                        "{} {} {}",
                        msg,
                        util.displayable_path(original),
                        util.displayable_path(converted),
                    )
                else:
                    # No transcoding necessary.
                    msg = (
                        "Hardlinking"
                        if hardlink
                        else ("Linking" if link else "Copying")
                    )

                    self._log.info("{} {.filepath}", msg, item)

                    if hardlink:
                        util.hardlink(original, converted)
                    elif link:
                        util.link(original, converted)
                    else:
                        util.copy(original, converted)

            if pretend:
                continue

            id3v23 = self.config["id3v23"].as_choice([True, False, "inherit"])
            if id3v23 == "inherit":
                id3v23 = None

            # Write tags from the database to the file if requested
            if self.config["write_metadata"].get(bool):
                item.try_write(path=converted, id3v23=id3v23)

            if keep_new:
                # If we're keeping the transcoded file, read it again (after
                # writing) to get new bitrate, duration, etc.
                item.path = converted
                item.read()
                item.store()  # Store new path and audio data.

            if self.config["embed"] and not linked:
                album = item._cached_album
                if album and album.artpath:
                    maxwidth = self._get_art_resize(album.artpath)
                    self._log.debug(
                        "embedding album art from {.art_filepath}", album
                    )
                    art.embed_item(
                        self._log,
                        item,
                        album.artpath,
                        maxwidth,
                        itempath=converted,
                        id3v23=id3v23,
                    )

            if keep_new:
                plugins.send(
                    "after_convert", item=item, dest=dest, keepnew=True
                )
            else:
                plugins.send(
                    "after_convert", item=item, dest=converted, keepnew=False
                )

    def copy_album_art(
        self,
        album,
        dest_dir,
        path_formats,
        pretend=False,
        link=False,
        hardlink=False,
    ):
        """Copies or converts the associated cover art of the album. Album must
        have at least one track.
        """
        if not album or not album.artpath:
            return

        album_item = album.items().get()
        # Album shouldn't be empty.
        if not album_item:
            return

        # Get the destination of the first item (track) of the album, we use
        # this function to format the path accordingly to path_formats.
        dest = album_item.destination(
            basedir=dest_dir, path_formats=path_formats
        )

        # Remove item from the path.
        dest = os.path.join(*util.components(dest)[:-1])

        dest = album.art_destination(album.artpath, item_dir=dest)
        if album.artpath == dest:
            return

        if not pretend:
            util.mkdirall(dest)

        if os.path.exists(util.syspath(dest)):
            self._log.info(
                "Skipping {.art_filepath} (target file exists)", album
            )
            return

        # Decide whether we need to resize the cover-art image.
        maxwidth = self._get_art_resize(album.artpath)

        # Either copy or resize (while copying) the image.
        if maxwidth is not None:
            self._log.info(
                "Resizing cover art from {.art_filepath} to {}",
                album,
                util.displayable_path(dest),
            )
            if not pretend:
                ArtResizer.shared.resize(maxwidth, album.artpath, dest)
        else:
            if pretend:
                msg = "ln" if hardlink else ("ln -s" if link else "cp")

                self._log.info(
                    "{} {.art_filepath} {}",
                    msg,
                    album,
                    util.displayable_path(dest),
                )
            else:
                msg = (
                    "Hardlinking"
                    if hardlink
                    else ("Linking" if link else "Copying")
                )

                self._log.info(
                    "{} cover art from {.art_filepath} to {}",
                    msg,
                    album,
                    util.displayable_path(dest),
                )
                if hardlink:
                    util.hardlink(album.artpath, dest)
                elif link:
                    util.link(album.artpath, dest)
                else:
                    util.copy(album.artpath, dest)

    def convert_func(self, lib, opts, args):
        (
            dest,
            threads,
            path_formats,
            fmt,
            pretend,
            hardlink,
            link,
            playlist,
            force,
        ) = self._get_opts_and_config(opts)

        if opts.album:
            albums = lib.albums(args)
            items = [i for a in albums for i in a.items()]
            if not pretend:
                for a in albums:
                    ui.print_(format(a, ""))
        else:
            items = list(lib.items(args))
            if not pretend:
                for i in items:
                    ui.print_(format(i, ""))

        if not items:
            self._log.error("Empty query result.")
            return
        if not (pretend or opts.yes or ui.input_yn("Convert? (Y/n)")):
            return

        if opts.album and self.config["copy_album_art"]:
            for album in albums:
                self.copy_album_art(
                    album, dest, path_formats, pretend, link, hardlink
                )

        # If the user supplied a playlist name, create a playlist for files
        # copied to the destination.
        pl_normpath = None
        items_paths = None
        if playlist:
            _, ext = get_format(fmt)
            # Playlist paths are understood as relative to the dest directory.
            pl_normpath = util.normpath(playlist)
            pl_dir = os.path.dirname(pl_normpath)
            items_paths = []
            for item in items:
                item_path = item.destination(
                    basedir=dest, path_formats=path_formats
                )

                # When keeping new files in the library, destination paths
                # keep original files and extensions.
                if not opts.keep_new and should_transcode(item, fmt, force):
                    item_path = replace_ext(item_path, ext)

                items_paths.append(os.path.relpath(item_path, pl_dir))

        self._parallel_convert(
            dest,
            opts.keep_new,
            path_formats,
            fmt,
            pretend,
            link,
            hardlink,
            threads,
            items,
            force,
        )

        if playlist:
            self._log.info("Creating playlist file {}", pl_normpath)
            if not pretend:
                m3ufile = M3UFile(playlist)
                m3ufile.set_contents(items_paths)
                m3ufile.write()

    def convert_on_import(self, lib, item):
        """Transcode a file automatically after it is imported into the
        library.
        """
        fmt = self.config["format"].as_str().lower()
        if should_transcode(item, fmt):
            command, ext = get_format()

            # Create a temporary file for the conversion.
            tmpdir = self.config["tmpdir"].get()
            if tmpdir:
                tmpdir = os.fsdecode(util.bytestring_path(tmpdir))
            fd, dest = tempfile.mkstemp(f".{os.fsdecode(ext)}", dir=tmpdir)
            os.close(fd)
            dest = util.bytestring_path(dest)
            _temp_files.append(dest)  # Delete the transcode later.

            # Convert.
            try:
                self.encode(command, item.path, dest)
            except subprocess.CalledProcessError:
                return

            # Change the newly-imported database entry to point to the
            # converted file.
            source_path = item.path
            item.path = dest
            item.write()
            item.read()  # Load new audio information data.
            item.store()

            if self.config["delete_originals"]:
                self._log.log(
                    logging.DEBUG if self.config["quiet"] else logging.INFO,
                    "Removing original file {}",
                    source_path,
                )
                util.remove(source_path, False)

    def _get_art_resize(self, artpath):
        """For a given piece of album art, determine whether or not it needs
        to be resized according to the user's settings. If so, returns the
        new size. If not, returns None.
        """
        newwidth = None
        if self.config["album_art_maxwidth"]:
            maxwidth = self.config["album_art_maxwidth"].get(int)
            size = ArtResizer.shared.get_size(artpath)
            self._log.debug("image size: {}", size)
            if size:
                if size[0] > maxwidth:
                    newwidth = maxwidth
            else:
                self._log.warning(
                    "Could not get size of image (please see "
                    "documentation for dependencies)."
                )
        return newwidth

    def _cleanup(self, task, session):
        for path in task.old_paths:
            if path in _temp_files:
                if os.path.isfile(util.syspath(path)):
                    util.remove(path)
                _temp_files.remove(path)

    def _get_opts_and_config(self, opts):
        """Returns parameters needed for convert function.
        Get parameters from command line if available,
        default to config if not available.
        """
        dest = opts.dest or self.config["dest"].get()
        if not dest:
            raise ui.UserError("no convert destination set")
        dest = util.bytestring_path(dest)

        threads = opts.threads or self.config["threads"].get(int)

        path_formats = ui.get_path_formats(self.config["paths"] or None)

        fmt = opts.format or self.config["format"].as_str().lower()

        playlist = opts.playlist or self.config["playlist"].get()
        if playlist is not None:
            playlist = os.path.join(dest, util.bytestring_path(playlist))

        if opts.pretend is not None:
            pretend = opts.pretend
        else:
            pretend = self.config["pretend"].get(bool)

        if opts.hardlink is not None:
            hardlink = opts.hardlink
            link = False
        elif opts.link is not None:
            hardlink = False
            link = opts.link
        else:
            hardlink = self.config["hardlink"].get(bool)
            link = self.config["link"].get(bool)
        force = getattr(opts, "force", False)
        return (
            dest,
            threads,
            path_formats,
            fmt,
            pretend,
            hardlink,
            link,
            playlist,
            force,
        )

    def _parallel_convert(
        self,
        dest,
        keep_new,
        path_formats,
        fmt,
        pretend,
        link,
        hardlink,
        threads,
        items,
        force,
    ):
        """Run the convert_item function for every items on as many thread as
        defined in threads
        """
        convert = [
            self.convert_item(
                dest,
                keep_new,
                path_formats,
                fmt,
                pretend,
                link,
                hardlink,
                force,
            )
            for _ in range(threads)
        ]
        pipe = util.pipeline.Pipeline([iter(items), convert])
        pipe.run_parallel()



from unittest.mock import MagicMock

import pytest
import requests

from beetsplug._utils.requests import RateLimitAdapter


def _prepared_request(
    url: str = "https://example.com",
) -> requests.PreparedRequest:
    req = requests.Request("GET", url)
    return req.prepare()


class TestRateLimitAdapter:
    @pytest.mark.parametrize(
        "last_request_time, now, expected_sleep",
        [
            (100.0, 100.0, 0.25),
            (100.0, 100.1, 0.15),
        ],
    )
    def test_send_sleeps_for_remaining_time(
        self, monkeypatch, last_request_time, now, expected_sleep
    ):
        adapter = RateLimitAdapter(rate_limit=0.25)
        request = _prepared_request()

        send_mock = MagicMock(return_value="ok")
        monkeypatch.setattr(
            "beetsplug._utils.requests.HTTPAdapter.send", send_mock
        )

        monkeypatch.setattr(
            "beetsplug._utils.requests.time.monotonic",
            lambda: now,
        )

        sleep_mock = MagicMock()
        monkeypatch.setattr("beetsplug._utils.requests.time.sleep", sleep_mock)

        adapter._last_request_time = last_request_time
        adapter.send(request)

        assert sleep_mock.call_count == 1
        assert sleep_mock.call_args.args[0] == pytest.approx(expected_sleep)



"""Updates the Emby Library whenever the beets library is changed.

emby:
    host: localhost
    port: 8096
    username: user
    apikey: apikey
    password: password
"""

import hashlib
from urllib.parse import parse_qs, urlencode, urljoin, urlsplit, urlunsplit

import requests

from beets.plugins import BeetsPlugin


def api_url(host, port, endpoint):
    """Returns a joined url.

    Takes host, port and endpoint and generates a valid emby API url.

    :param host: Hostname of the emby server
    :param port: Portnumber of the emby server
    :param endpoint: API endpoint
    :type host: str
    :type port: int
    :type endpoint: str
    :returns: Full API url
    :rtype: str
    """
    # check if http or https is defined as host and create hostname
    hostname_list = [host]
    if host.startswith("http://") or host.startswith("https://"):
        hostname = "".join(hostname_list)
    else:
        hostname_list.insert(0, "http://")
        hostname = "".join(hostname_list)

    joined = urljoin(f"{hostname}:{port}", endpoint)

    scheme, netloc, path, query_string, fragment = urlsplit(joined)
    query_params = parse_qs(query_string)

    query_params["format"] = ["json"]
    new_query_string = urlencode(query_params, doseq=True)

    return urlunsplit((scheme, netloc, path, new_query_string, fragment))


def password_data(username, password):
    """Returns a dict with username and its encoded password.

    :param username: Emby username
    :param password: Emby password
    :type username: str
    :type password: str
    :returns: Dictionary with username and encoded password
    :rtype: dict
    """
    return {
        "username": username,
        "password": hashlib.sha1(password.encode("utf-8")).hexdigest(),
        "passwordMd5": hashlib.md5(password.encode("utf-8")).hexdigest(),
    }


def create_headers(user_id, token=None):
    """Return header dict that is needed to talk to the Emby API.

    :param user_id: Emby user ID
    :param token: Authentication token for Emby
    :type user_id: str
    :type token: str
    :returns: Headers for requests
    :rtype: dict
    """
    headers = {}

    authorization = (
        f'MediaBrowser UserId="{user_id}", '
        'Client="other", '
        'Device="beets", '
        'DeviceId="beets", '
        'Version="0.0.0"'
    )

    headers["x-emby-authorization"] = authorization

    if token:
        headers["x-mediabrowser-token"] = token

    return headers


def get_token(host, port, headers, auth_data):
    """Return token for a user.

    :param host: Emby host
    :param port: Emby port
    :param headers: Headers for requests
    :param auth_data: Username and encoded password for authentication
    :type host: str
    :type port: int
    :type headers: dict
    :type auth_data: dict
    :returns: Access Token
    :rtype: str
    """
    url = api_url(host, port, "/Users/AuthenticateByName")
    r = requests.post(
        url,
        headers=headers,
        data=auth_data,
        timeout=10,
    )

    return r.json().get("AccessToken")


def get_user(host, port, username):
    """Return user dict from server or None if there is no user.

    :param host: Emby host
    :param port: Emby port
    :username: Username
    :type host: str
    :type port: int
    :type username: str
    :returns: Matched Users
    :rtype: list
    """
    url = api_url(host, port, "/Users/Public")
    r = requests.get(url, timeout=10)
    user = [i for i in r.json() if i["Name"] == username]

    return user


class EmbyUpdate(BeetsPlugin):
    def __init__(self):
        super().__init__("emby")

        # Adding defaults.
        self.config.add(
            {
                "host": "http://localhost",
                "port": 8096,
                "username": None,
                "password": None,
                "userid": None,
                "apikey": None,
            }
        )
        self.config["username"].redact = True
        self.config["password"].redact = True
        self.config["userid"].redact = True
        self.config["apikey"].redact = True

        self.register_listener("database_change", self.listen_for_db_change)

    def listen_for_db_change(self, lib, model):
        """Listens for beets db change and register the update for the end."""
        self.register_listener("cli_exit", self.update)

    def update(self, lib):
        """When the client exists try to send refresh request to Emby."""
        self._log.info("Updating Emby library...")

        host = self.config["host"].get()
        port = self.config["port"].get()
        username = self.config["username"].get()
        password = self.config["password"].get()
        userid = self.config["userid"].get()
        token = self.config["apikey"].get()

        # Check if at least a apikey or password is given.
        if not any([password, token]):
            self._log.warning("Provide at least Emby password or apikey.")
            return

        if not userid:
            # Get user information from the Emby API.
            user = get_user(host, port, username)
            if not user:
                self._log.warning("User {} could not be found.", username)
                return
            userid = user[0]["Id"]

        if not token:
            # Create Authentication data and headers.
            auth_data = password_data(username, password)
            headers = create_headers(userid)

            # Get authentication token.
            token = get_token(host, port, headers, auth_data)
            if not token:
                self._log.warning("Could not get token for user {}", username)
                return

        # Recreate headers with a token.
        headers = create_headers(userid, token=token)

        # Trigger the Update.
        url = api_url(host, port, "/Library/Refresh")
        r = requests.post(
            url,
            headers=headers,
            timeout=10,
        )
        if r.status_code != 204:
            self._log.warning("Update could not be triggered")
        else:
            self._log.info("Update triggered.")



# This file is part of beets.
# Copyright 2016, Adrian Sampson.
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.

"""Fetches, embeds, and displays lyrics."""

from __future__ import annotations

import itertools
import math
import re
import textwrap
from contextlib import contextmanager, suppress
from dataclasses import dataclass
from functools import cached_property, partial, total_ordering
from html import unescape
from itertools import filterfalse, groupby
from pathlib import Path
from typing import TYPE_CHECKING, ClassVar, NamedTuple
from urllib.parse import quote, quote_plus, urlencode, urlparse

import requests
from bs4 import BeautifulSoup
from unidecode import unidecode

from beets import plugins, ui
from beets.autotag.distance import string_dist
from beets.dbcore import types
from beets.dbcore.query import FalseQuery
from beets.library import Item, parse_query_string
from beets.util.config import sanitize_choices
from beets.util.lyrics import INSTRUMENTAL_LYRICS, Lyrics

from ._utils.requests import HTTPNotFoundError, RequestHandler

if TYPE_CHECKING:
    from collections.abc import Iterable, Iterator

    import confuse

    from beets.importer import ImportTask
    from beets.library import Library
    from beets.logging import BeetsLogger as Logger

    from ._typing import (
        GeniusAPI,
        GoogleCustomSearchAPI,
        JSONDict,
        LRCLibAPI,
        TranslatorAPI,
    )


class CaptchaError(requests.exceptions.HTTPError):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__("Captcha is required", *args, **kwargs)


class GeniusHTTPError(requests.exceptions.HTTPError):
    pass


# Utilities.


def search_pairs(item):
    """Yield a pairs of artists and titles to search for.

    The first item in the pair is the name of the artist, the second
    item is a list of song names.

    In addition to the artist and title obtained from the `item` the
    method tries to strip extra information like paranthesized suffixes
    and featured artists from the strings and add them as candidates.
    The artist sort name is added as a fallback candidate to help in
    cases where artist name includes special characters or is in a
    non-latin script.
    The method also tries to split multiple titles separated with `/`.
    """

    def generate_alternatives(string, patterns):
        """Generate string alternatives by extracting first matching group for
        each given pattern.
        """
        alternatives = [string]
        for pattern in patterns:
            match = re.search(pattern, string, re.IGNORECASE)
            if match:
                alternatives.append(match.group(1))
        return alternatives

    title, artist, artist_sort = (
        item.title.strip(),
        item.artist.strip(),
        item.artist_sort.strip(),
    )
    if not title or not artist:
        return ()

    patterns = [
        # Remove any featuring artists from the artists name
        rf"(.*?) {plugins.feat_tokens()}"
    ]

    # Skip various artists
    artists = []
    lower_artist = artist.lower()
    if "various" not in lower_artist:
        artists.extend(generate_alternatives(artist, patterns))
    # Use the artist_sort as fallback only if it differs from artist to avoid
    # repeated remote requests with the same search terms
    artist_sort_lower = artist_sort.lower()
    if (
        artist_sort
        and lower_artist != artist_sort_lower
        and "various" not in artist_sort_lower
    ):
        artists.append(artist_sort)

    patterns = [
        # Remove a parenthesized suffix from a title string. Common
        # examples include (live), (remix), and (acoustic).
        r"(.+?)\s+[(].*[)]$",
        # Remove any featuring artists from the title
        rf"(.*?) {plugins.feat_tokens(for_artist=False)}",
        # Remove part of title after colon ':' for songs with subtitles
        r"(.+?)\s*:.*",
    ]
    titles = generate_alternatives(title, patterns)

    # Check for a dual song (e.g. Pink Floyd - Speak to Me / Breathe)
    # and each of them.
    multi_titles = []
    for title in titles:
        multi_titles.append([title])
        if " / " in title:
            multi_titles.append([x.strip() for x in title.split(" / ")])

    return itertools.product(artists, multi_titles)


def slug(text: str) -> str:
    """Make a URL-safe, human-readable version of the given text

    This will do the following:

    1. decode unicode characters into ASCII
    2. shift everything to lowercase
    3. strip whitespace
    4. replace other non-word characters with dashes
    5. strip extra dashes
    """
    return re.sub(r"\W+", "-", unidecode(text).lower().strip()).strip("-")


class LyricsRequestHandler(RequestHandler):
    _log: Logger

    def status_to_error(self, code: int) -> type[requests.HTTPError] | None:
        if err := super().status_to_error(code):
            return err

        if 300 <= code < 400:
            return CaptchaError

        return None

    def debug(self, message: str, *args) -> None:
        """Log a debug message with the class name."""
        self._log.debug(f"{self.__class__.__name__}: {message}", *args)

    def info(self, message: str, *args) -> None:
        """Log an info message with the class name."""
        self._log.info(f"{self.__class__.__name__}: {message}", *args)

    def warn(self, message: str, *args) -> None:
        """Log warning with the class name."""
        self._log.warning(f"{self.__class__.__name__}: {message}", *args)

    @staticmethod
    def format_url(url: str, params: JSONDict | None) -> str:
        if not params:
            return url

        return f"{url}?{urlencode(params)}"

    def get_text(
        self, url: str, params: JSONDict | None = None, **kwargs
    ) -> str:
        """Return text / HTML data from the given URL.

        Set the encoding to None to let requests handle it because some sites
        set it incorrectly.
        """
        url = self.format_url(url, params)
        self.debug("Fetching HTML from {}", url)
        r = self.get(url, **kwargs)
        r.encoding = None
        return r.text

    def get_json(self, url: str, params: JSONDict | None = None, **kwargs):
        """Return JSON data from the given URL."""
        url = self.format_url(url, params)
        self.debug("Fetching JSON from {}", url)
        return super().get_json(url, **kwargs)

    def post_json(self, url: str, params: JSONDict | None = None, **kwargs):
        """Send POST request and return JSON response."""
        url = self.format_url(url, params)
        self.debug("Posting JSON to {}", url)
        return self.request("post", url, **kwargs).json()

    @contextmanager
    def handle_request(self) -> Iterator[None]:
        try:
            yield
        except requests.JSONDecodeError:
            self.warn("Could not decode response JSON data")
        except requests.RequestException as exc:
            self.warn("Request error: {}", exc)


class BackendClass(type):
    @property
    def name(cls) -> str:
        """Return lowercase name of the backend class."""
        return cls.__name__.lower()


class Backend(LyricsRequestHandler, metaclass=BackendClass):
    config: confuse.Subview

    def __init__(self, config: confuse.Subview, log: Logger) -> None:
        self._log = log
        self.config = config

    def fetch(
        self, artist: str, title: str, album: str, length: int
    ) -> Lyrics | None:
        """Return lyrics for a song, or ``None`` when no match is found."""
        raise NotImplementedError


@dataclass
@total_ordering
class LRCLyrics:
    """Hold LRCLib candidate data and ranking helpers for matching."""

    #: Percentage tolerance for max duration difference between lyrics and item.
    DURATION_DIFF_TOLERANCE = 0.05

    target_duration: float
    id: int
    duration: float
    instrumental: bool
    plain: str
    synced: str | None

    def __le__(self, other: LRCLyrics) -> bool:
        """Compare two lyrics items by their score."""
        return self.dist < other.dist

    @classmethod
    def verify_synced_lyrics(
        cls, duration: float, lyrics: str | None
    ) -> str | None:
        """Accept synced lyrics only when the final timestamp fits duration."""
        if lyrics and (
            m := Lyrics.LINE_PARTS_PAT.match(lyrics.splitlines()[-1])
        ):
            ts, _ = m.groups()
            if ts:
                mm, ss = map(float, ts.strip("[]").split(":"))
                if 60 * mm + ss <= duration:
                    return lyrics

        return None

    @classmethod
    def make(
        cls, candidate: LRCLibAPI.Item, target_duration: float
    ) -> LRCLyrics:
        """Build a scored candidate from LRCLib payload data."""
        duration = candidate["duration"] or 0.0
        return cls(
            target_duration,
            candidate["id"],
            duration,
            candidate["instrumental"],
            candidate["plainLyrics"],
            cls.verify_synced_lyrics(
                target_duration, candidate["syncedLyrics"]
            ),
        )

    @cached_property
    def duration_dist(self) -> float:
        """Return the absolute difference between lyrics and target duration."""
        return abs(self.duration - self.target_duration)

    @cached_property
    def is_valid(self) -> bool:
        """Return whether the lyrics item is valid.
        Lyrics duration must be within the tolerance defined by
        :attr:`DURATION_DIFF_TOLERANCE`.
        """
        return (
            self.duration_dist
            <= self.target_duration * self.DURATION_DIFF_TOLERANCE
        )

    @cached_property
    def dist(self) -> tuple[bool, float]:
        """Distance/score of the given lyrics item.

        Return a tuple with the following values:
        1. Absolute difference between lyrics and target duration
        2. Boolean telling whether synced lyrics are available.

        Best lyrics match is the one that has the closest duration to
        ``target_duration`` and has synced lyrics available.
        """
        return not self.synced, self.duration_dist

    def get_text(self, want_synced: bool) -> str:
        """Return the preferred text form for this candidate."""
        if self.instrumental:
            return INSTRUMENTAL_LYRICS

        if want_synced and self.synced:
            return "\n".join(map(str.strip, self.synced.splitlines()))

        return self.plain


class LRCLib(Backend):
    """Fetch lyrics from the LRCLib API."""

    BASE_URL = "https://lrclib.net/api"
    GET_URL = f"{BASE_URL}/get"
    SEARCH_URL = f"{BASE_URL}/search"

    def fetch_candidates(
        self, artist: str, title: str, album: str, length: int
    ) -> Iterator[list[LRCLibAPI.Item]]:
        """Yield lyrics candidates for the given song data.

        I found that the ``/get`` endpoint sometimes returns inaccurate or
        unsynced lyrics, while ``search`` yields more suitable candidates.
        Therefore, we prioritize the latter and rank the results using our own
        algorithm. If the search does not give suitable lyrics, we fall back to
        the ``/get`` endpoint.

        Return an iterator over lists of candidates.
        """
        base_params = {"artist_name": artist, "track_name": title}
        get_params = {**base_params, "duration": length}
        if album:
            get_params["album_name"] = album

        yield self.get_json(self.SEARCH_URL, params=base_params)

        with suppress(HTTPNotFoundError):
            yield [self.get_json(self.GET_URL, params=get_params)]

    @classmethod
    def pick_best_match(cls, lyrics: Iterable[LRCLyrics]) -> LRCLyrics | None:
        """Return best matching lyrics item from the given list."""
        return min((li for li in lyrics if li.is_valid), default=None)

    def fetch(
        self, artist: str, title: str, album: str, length: int
    ) -> Lyrics | None:
        """Fetch lyrics text for the given song data."""
        evaluate_item = partial(LRCLyrics.make, target_duration=length)

        for group in self.fetch_candidates(artist, title, album, length):
            candidates = [evaluate_item(item) for item in group]
            if item := self.pick_best_match(candidates):
                lyrics = item.get_text(self.config["synced"].get(bool))
                return Lyrics(
                    lyrics, self.__class__.name, f"{self.GET_URL}/{item.id}"
                )

        return None


class MusiXmatch(Backend):
    URL_TEMPLATE = "https://www.musixmatch.com/lyrics/{}/{}"

    REPLACEMENTS: ClassVar[dict[str, str]] = {
        "['\u2018\u2019]": "-",
        r"\s+": "-",
        "<": "Less_Than",
        ">": "Greater_Than",
        "#": "Number_",
        r"[\[\{]": "(",
        r"[\]\}]": ")",
    }

    @classmethod
    def encode(cls, text: str) -> str:
        for old, new in cls.REPLACEMENTS.items():
            text = re.sub(old, new, text)

        return quote(unidecode(text))

    @classmethod
    def build_url(cls, *args: str) -> str:
        return cls.URL_TEMPLATE.format(*map(cls.encode, args))

    def fetch(self, artist: str, title: str, *_) -> Lyrics | None:
        url = self.build_url(artist, title)

        html = self.get_text(url)
        if "We detected that your IP is blocked" in html:
            self.warn("Failed: Blocked IP address")
            return None
        html_parts = html.split('<p class="mxm-lyrics__content')
        # Sometimes lyrics come in 2 or more parts
        lyrics_parts = []
        for html_part in html_parts:
            lyrics_parts.append(re.sub(r"^[^>]+>|</p>.*", "", html_part))
        lyrics = "\n".join(lyrics_parts)
        lyrics = lyrics.strip(',"').replace("\\n", "\n")
        # another odd case: sometimes only that string remains, for
        # missing songs. this seems to happen after being blocked
        # above, when filling in the CAPTCHA.
        if "Instant lyrics for all your music." in lyrics:
            return None
        # sometimes there are non-existent lyrics with some content
        if "Lyrics | Musixmatch" in lyrics:
            return None
        return Lyrics(lyrics, self.__class__.name, url)


class Html:
    collapse_space = partial(re.compile(r"(^| ) +", re.M).sub, r"\1")
    expand_br = partial(re.compile(r"\s*<br[^>]*>\s*", re.I).sub, "\n")
    #: two newlines between paragraphs on the same line (musica, letras.mus.br)
    merge_blocks = partial(re.compile(r"(?<!>)</p><p[^>]*>").sub, "\n\n")
    #: a single new line between paragraphs on separate lines
    #: (paroles.net, sweetslyrics.com, lacoccinelle.net)
    merge_lines = partial(re.compile(r"</p>\s+<p[^>]*>(?!___)").sub, "\n")
    #: remove empty divs (lacoccinelle.net)
    remove_empty_tags = partial(
        re.compile(r"(<(div|span)[^>]*>\s*</\2>)").sub, ""
    )
    #: remove Google Ads tags (musica.com)
    remove_aside = partial(re.compile("<aside .+?</aside>").sub, "")
    #: remove adslot-Content_1 div from the lyrics text (paroles.net)
    remove_adslot = partial(
        re.compile(r"\n</div>[^\n]+-- Content_\d+ --.*?\n<div>", re.S).sub,
        "\n",
    )
    #: remove text formatting (azlyrics.com, lacocinelle.net)
    remove_formatting = partial(
        re.compile(r" *</?(i|em|pre|strong)[^>]*>").sub, ""
    )

    @classmethod
    def normalize_space(cls, text: str) -> str:
        text = unescape(text).replace("\r", "").replace("\xa0", " ")
        return cls.collapse_space(cls.expand_br(text))

    @classmethod
    def remove_ads(cls, text: str) -> str:
        return cls.remove_adslot(cls.remove_aside(text))

    @classmethod
    def merge_paragraphs(cls, text: str) -> str:
        return cls.merge_blocks(cls.merge_lines(cls.remove_empty_tags(text)))


class SoupMixin:
    @classmethod
    def pre_process_html(cls, html: str) -> str:
        """Pre-process the HTML content before scraping."""
        return Html.normalize_space(html)

    @classmethod
    def get_soup(cls, html: str) -> BeautifulSoup:
        return BeautifulSoup(cls.pre_process_html(html), "html.parser")


class SearchResult(NamedTuple):
    artist: str
    title: str
    url: str

    @property
    def source(self) -> str:
        return urlparse(self.url).netloc


class SearchBackend(SoupMixin, Backend):
    @cached_property
    def dist_thresh(self) -> float:
        return self.config["dist_thresh"].get(float)

    def check_match(
        self, target_artist: str, target_title: str, result: SearchResult
    ) -> bool:
        """Check if the given search result is a 'good enough' match."""
        max_dist = max(
            string_dist(target_artist, result.artist),
            string_dist(target_title, result.title),
        )

        if (max_dist := round(max_dist, 2)) <= self.dist_thresh:
            return True

        if math.isclose(max_dist, self.dist_thresh, abs_tol=0.4):
            # log out the candidate that did not make it but was close.
            # This may show a matching candidate with some noise in the name
            self.debug(
                "({0.artist}, {0.title}) does not match ({1}, {2}) but dist"
                " was close: {3:.2f}",
                result,
                target_artist,
                target_title,
                max_dist,
            )

        return False

    def search(self, artist: str, title: str) -> Iterable[SearchResult]:
        """Search for the given query and yield search results."""
        raise NotImplementedError

    def get_results(self, artist: str, title: str) -> Iterable[SearchResult]:
        check_match = partial(self.check_match, artist, title)
        for candidate in self.search(artist, title):
            if check_match(candidate):
                yield candidate

    def fetch(self, artist: str, title: str, *_) -> Lyrics | None:
        """Fetch lyrics for the given artist and title."""
        for result in self.get_results(artist, title):
            if (html := self.get_text(result.url)) and (
                lyrics := self.scrape(html)
            ):
                return Lyrics(lyrics, self.__class__.name, result.url)

        return None

    @classmethod
    def scrape(cls, html: str) -> str | None:
        """Scrape the lyrics from the given HTML."""
        raise NotImplementedError


class Genius(SearchBackend):
    """Fetch lyrics from Genius via genius-api.

    Because genius doesn't allow accessing lyrics via the api, we first query
    the api for a url matching our artist & title, then scrape the HTML text
    for the JSON data containing the lyrics.
    """

    SEARCH_URL = "https://api.genius.com/search"
    LYRICS_IN_JSON_RE = re.compile(r'(?<=.\\"html\\":\\").*?(?=(?<!\\)\\")')
    remove_backslash = partial(re.compile(r"\\(?=[^\\])").sub, "")

    @cached_property
    def headers(self) -> dict[str, str]:
        return {"Authorization": f"Bearer {self.config['genius_api_key']}"}

    def get_json(self, *args, **kwargs) -> GeniusAPI.Search:
        response: GeniusAPI.Response = super().get_json(*args, **kwargs)
        if "response" in response:
            return response  # type: ignore[return-value]

        meta = response["meta"]
        raise GeniusHTTPError(f"{meta['message']} Status: {meta['status']}")

    def search(self, artist: str, title: str) -> Iterable[SearchResult]:
        search_data = self.get_json(
            self.SEARCH_URL,
            params={"q": f"{artist} {title}"},
            headers=self.headers,
        )
        for r in (hit["result"] for hit in search_data["response"]["hits"]):
            yield SearchResult(r["artist_names"], r["title"], r["url"])

    @classmethod
    def scrape(cls, html: str) -> str | None:
        if m := cls.LYRICS_IN_JSON_RE.search(html):
            html_text = cls.remove_backslash(m[0]).replace(r"\n", "\n")
            return cls.get_soup(html_text).get_text().strip()

        return None


class Tekstowo(SearchBackend):
    """Fetch lyrics from Tekstowo.pl."""

    BASE_URL = "https://www.tekstowo.pl"
    SEARCH_URL = f"{BASE_URL}/szukaj,{{}}.html"

    def build_url(self, artist, title):
        artistitle = f"{artist.title()} {title.title()}"

        return self.SEARCH_URL.format(quote_plus(unidecode(artistitle)))

    def search(self, artist: str, title: str) -> Iterable[SearchResult]:
        if html := self.get_text(self.build_url(title, artist)):
            soup = self.get_soup(html)
            for tag in soup.select("div[class=flex-group] > a[title*=' - ']"):
                artist, title = str(tag["title"]).split(" - ", 1)
                yield SearchResult(
                    artist, title, f"{self.BASE_URL}{tag['href']}"
                )

        return None

    @classmethod
    def scrape(cls, html: str) -> str | None:
        soup = cls.get_soup(html)

        if lyrics_div := soup.select_one("div.song-text > div.inner-text"):
            return lyrics_div.get_text()

        return None


class Google(SearchBackend):
    """Fetch lyrics from Google search results."""

    SEARCH_URL = "https://www.googleapis.com/customsearch/v1"

    #: Exclude some letras.mus.br pages which do not contain lyrics.
    EXCLUDE_PAGES: ClassVar[list[str]] = [
        "significado.html",
        "traduccion.html",
        "traducao.html",
        "significados.html",
    ]

    #: Regular expression to match noise in the URL title.
    URL_TITLE_NOISE_RE = re.compile(
        r"""
\b
(
      paroles(\ et\ traduction|\ de\ chanson)?
    | letras?(\ de)?
    | liedtexte
    | dainų\ žodžiai
    | original\ song\ full\ text\.
    | official
    | 20[12]\d\ version
    | (absolute\ |az)?lyrics(\ complete)?
    | www\S+
    | \S+\.(com|net|mus\.br)
)
([^\w.]|$)
""",
        re.IGNORECASE | re.VERBOSE,
    )
    #: Split cleaned up URL title into artist and title parts.
    URL_TITLE_PARTS_RE = re.compile(r" +(?:[ :|-]+|par|by) +|, ")

    SOURCE_DIST_FACTOR: ClassVar[dict[str, float]] = {
        "www.azlyrics.com": 0.5,
        "www.songlyrics.com": 0.6,
    }

    ignored_domains: ClassVar[set[str]] = set()

    @classmethod
    def pre_process_html(cls, html: str) -> str:
        """Pre-process the HTML content before scraping."""
        html = Html.remove_ads(super().pre_process_html(html))
        return Html.remove_formatting(Html.merge_paragraphs(html))

    def get_text(self, *args, **kwargs) -> str:
        """Handle an error so that we can continue with the next URL."""
        kwargs.setdefault("allow_redirects", False)
        with self.handle_request():
            try:
                return super().get_text(*args, **kwargs)
            except CaptchaError:
                self.ignored_domains.add(urlparse(args[0]).netloc)
                raise

    @staticmethod
    def get_part_dist(artist: str, title: str, part: str) -> float:
        """Return the distance between the given part and the artist and title.

        A number between -1 and 1 is returned, where -1 means the part is
        closer to the artist and 1 means it is closer to the title.
        """
        return string_dist(artist, part) - string_dist(title, part)

    @classmethod
    def make_search_result(
        cls, artist: str, title: str, item: GoogleCustomSearchAPI.Item
    ) -> SearchResult:
        """Parse artist and title from the URL title and return a search result."""
        url_title = (
            # get full title from metatags if available
            item.get("pagemap", {}).get("metatags", [{}])[0].get("og:title")
            # default to the dispolay title
            or item["title"]
        )
        clean_title = cls.URL_TITLE_NOISE_RE.sub("", url_title).strip(" .-|")
        # split it into parts which may be part of the artist or the title
        # `dict.fromkeys` removes duplicates keeping the order
        parts = list(dict.fromkeys(cls.URL_TITLE_PARTS_RE.split(clean_title)))

        if len(parts) == 1:
            part = parts[0]
            if m := re.search(rf"(?i)\W*({re.escape(title)})\W*", part):
                # artist and title may not have a separator
                result_title = m[1]
                result_artist = part.replace(m[0], "")
            else:
                # assume that this is the title
                result_artist, result_title = "", parts[0]
        else:
            # sort parts by their similarity to the artist
            result_artist = min(parts, key=lambda p: string_dist(artist, p))
            result_title = min(parts, key=lambda p: string_dist(title, p))

        return SearchResult(result_artist, result_title, item["link"])

    def search(self, artist: str, title: str) -> Iterable[SearchResult]:
        params = {
            "key": self.config["google_API_key"].as_str(),
            "cx": self.config["google_engine_ID"].as_str(),
            "q": f"{artist} {title}",
            "siteSearch": "www.musixmatch.com",
            "siteSearchFilter": "e",
            "excludeTerms": ", ".join(self.EXCLUDE_PAGES),
        }

        data: GoogleCustomSearchAPI.Response = self.get_json(
            self.SEARCH_URL, params=params
        )
        for item in data.get("items", []):
            yield self.make_search_result(artist, title, item)

    def get_results(self, *args) -> Iterable[SearchResult]:
        """Try results from preferred sources first."""
        for result in sorted(
            super().get_results(*args),
            key=lambda r: self.SOURCE_DIST_FACTOR.get(r.source, 1),
        ):
            if result.source not in self.ignored_domains:
                yield result

    @classmethod
    def scrape(cls, html: str) -> str | None:
        # Get the longest text element (if any).
        if strings := sorted(cls.get_soup(html).stripped_strings, key=len):
            return strings[-1]

        return None


@dataclass
class Translator(LyricsRequestHandler):
    """Translate lyrics text while preserving existing structured metadata."""

    TRANSLATE_URL = "https://api.cognitive.microsofttranslator.com/translate"
    SEPARATOR = " | "

    _log: Logger
    api_key: str
    to_language: str
    from_languages: list[str]

    @classmethod
    def from_config(
        cls,
        log: Logger,
        api_key: str,
        to_language: str,
        from_languages: list[str] | None = None,
    ) -> Translator:
        """Construct a translator with normalized language configuration."""
        return cls(
            log,
            api_key,
            to_language.upper(),
            [x.upper() for x in from_languages or []],
        )

    def get_translations(self, texts: Iterable[str]) -> list[str]:
        """Return translations for the given texts.

        To reduce the translation 'cost', we translate unique texts, and then
        map the translations back to the original texts.
        """
        unique_texts = list(dict.fromkeys(texts))
        text = self.SEPARATOR.join(unique_texts)
        data: list[TranslatorAPI.Response] = self.post_json(
            self.TRANSLATE_URL,
            headers={"Ocp-Apim-Subscription-Key": self.api_key},
            json=[{"text": text}],
            params={"api-version": "3.0", "to": self.to_language},
        )

        translated_text = data[0]["translations"][0]["text"]
        translations = translated_text.split(self.SEPARATOR)
        trans_by_text = dict(zip(unique_texts, translations))
        return [trans_by_text.get(t, "") for t in texts]

    def translate(self, lyrics: Lyrics, old_lyrics: Lyrics) -> Lyrics:
        """Translate the given lyrics to the target language.

        Check old lyrics for existing translations and return them if their
        original text matches the new lyrics. This is to avoid translating
        the same lyrics multiple times.

        If the lyrics are already in the target language or not in any of
        of the source languages (if configured), they are returned as is.
        """
        if (
            lyrics.original_text
        ) == old_lyrics.original_text and old_lyrics.translated:
            self.info("🔵 Translations already exist")
            return old_lyrics

        if (lyrics_language := lyrics.language) == self.to_language:
            self.info(
                "🔵 Lyrics are already in the target language {.to_language}",
                self,
            )
        elif (
            from_lang_config := self.from_languages
        ) and lyrics_language not in from_lang_config:
            self.info(
                "🔵 Configuration {} does not permit translating from {}",
                from_lang_config,
                lyrics_language,
            )
        else:
            with self.handle_request():
                lyrics.translations = self.get_translations(lyrics.text_lines)
                lyrics.translation_language = self.to_language
                self.info("🟢 Translated lyrics to {.to_language}", self)

        return lyrics


@dataclass
class RestFiles:
    # The content for the base index.rst generated in ReST mode.
    REST_INDEX_TEMPLATE = textwrap.dedent("""
        Lyrics
        ======

        * :ref:`Song index <genindex>`
        * :ref:`search`

        Artist index:

        .. toctree::
           :maxdepth: 1
           :glob:

           artists/*
        """).strip()

    # The content for the base conf.py generated.
    REST_CONF_TEMPLATE = textwrap.dedent("""
        master_doc = "index"
        project = "Lyrics"
        copyright = "none"
        author = "Various Authors"
        latex_documents = [
            (master_doc, "Lyrics.tex", project, author, "manual"),
        ]
        epub_exclude_files = ["search.html"]
        epub_tocdepth = 1
        epub_tocdup = False
        """).strip()

    directory: Path

    @cached_property
    def artists_dir(self) -> Path:
        dir = self.directory / "artists"
        dir.mkdir(parents=True, exist_ok=True)
        return dir

    def write_indexes(self) -> None:
        """Write conf.py and index.rst files necessary for Sphinx

        We write minimal configurations that are necessary for Sphinx
        to operate. We do not overwrite existing files so that
        customizations are respected."""
        index_file = self.directory / "index.rst"
        if not index_file.exists():
            index_file.write_text(self.REST_INDEX_TEMPLATE)
        conf_file = self.directory / "conf.py"
        if not conf_file.exists():
            conf_file.write_text(self.REST_CONF_TEMPLATE)

    def write_artist(self, artist: str, items: Iterable[Item]) -> None:
        parts = [
            f"{artist}\n{'=' * len(artist)}",
            ".. contents::\n   :local:",
        ]
        for album, items in groupby(items, key=lambda i: i.album):
            parts.append(f"{album}\n{'-' * len(album)}")
            parts.extend(
                part
                for i in items
                if (title := f":index:`{i.title.strip()}`")
                for part in (
                    f"{title}\n{'~' * len(title)}",
                    textwrap.indent(i.lyrics, "| "),
                )
            )
        file = self.artists_dir / f"{slug(artist)}.rst"
        file.write_text("\n\n".join(parts).strip())

    def write(self, items: list[Item]) -> None:
        self.directory.mkdir(exist_ok=True, parents=True)
        self.write_indexes()

        items.sort(key=lambda i: i.albumartist)
        for artist, artist_items in groupby(items, key=lambda i: i.albumartist):
            self.write_artist(artist.strip(), artist_items)

        d = self.directory
        text = f"""
        ReST files generated. to build, use one of:
          sphinx-build -b html  {d} {d / "html"}
          sphinx-build -b epub  {d} {d / "epub"}
          sphinx-build -b latex {d} {d / "latex"} && make -C {d / "latex"} all-pdf
        """
        ui.print_(textwrap.dedent(text))


BACKEND_BY_NAME = {
    b.name: b for b in [LRCLib, Google, Genius, Tekstowo, MusiXmatch]
}


class LyricsPlugin(LyricsRequestHandler, plugins.BeetsPlugin):
    item_types: ClassVar[dict[str, types.Type]] = {
        "lyrics_url": types.STRING,
        "lyrics_backend": types.STRING,
        "lyrics_language": types.STRING,
        "lyrics_translation_language": types.STRING,
    }

    @cached_property
    def backends(self) -> list[Backend]:
        user_sources = self.config["sources"].as_str_seq()

        chosen = sanitize_choices(user_sources, BACKEND_BY_NAME)
        if "google" in chosen and not self.config["google_API_key"].get():
            self.warn("Disabling Google source: no API key configured.")
            chosen.remove("google")

        return [BACKEND_BY_NAME[c](self.config, self._log) for c in chosen]

    @cached_property
    def translator(self) -> Translator | None:
        config = self.config["translate"]
        if config["api_key"].get() and config["to_language"].get():
            return Translator.from_config(self._log, **config.flatten())
        return None

    def __init__(self):
        super().__init__()
        self.config.add(
            {
                "auto": True,
                "auto_ignore": None,
                "translate": {
                    "api_key": None,
                    "from_languages": [],
                    "to_language": None,
                },
                "dist_thresh": 0.11,
                "google_API_key": None,
                "google_engine_ID": "009217259823014548361:lndtuqkycfu",
                "genius_api_key": (
                    "Ryq93pUGm8bM6eUWwD_M3NOFFDAtp2yEE7W"
                    "76V-uFL5jks5dNvcGCdarqFjDhP9c"
                ),
                "fallback": None,
                "force": False,
                "keep_synced": False,
                "local": False,
                "print": False,
                "synced": False,
                # Musixmatch and Tekstowo are disabled by default as they
                # currently block requests with the beets user agent.
                "sources": [
                    n
                    for n in BACKEND_BY_NAME
                    if n not in {"musixmatch", "tekstowo"}
                ],
            }
        )
        self.config["translate"]["api_key"].redact = True
        self.config["google_API_key"].redact = True
        self.config["google_engine_ID"].redact = True
        self.config["genius_api_key"].redact = True

        if self.config["auto"]:
            self.import_stages = [self.imported]

    def commands(self):
        cmd = ui.Subcommand("lyrics", help="fetch song lyrics")
        cmd.parser.add_option(
            "-p",
            "--print",
            action="store_true",
            default=self.config["print"].get(),
            help="print lyrics to console",
        )
        cmd.parser.add_option(
            "-r",
            "--write-rest",
            dest="rest_directory",
            action="store",
            default=None,
            metavar="dir",
            help="write lyrics to given directory as ReST files",
        )
        cmd.parser.add_option(
            "-f",
            "--force",
            action="store_true",
            default=self.config["force"].get(),
            help="always re-download lyrics",
        )
        cmd.parser.add_option(
            "--keep-synced",
            action="store_true",
            default=self.config["keep_synced"].get(),
            help="re-download only unsynced lyrics",
        )
        cmd.parser.add_option(
            "-l",
            "--local",
            action="store_true",
            default=self.config["local"].get(),
            help="do not fetch missing lyrics",
        )

        def func(lib: Library, opts, args) -> None:
            # The "write to files" option corresponds to the
            # import_write config value.
            self.config.set(vars(opts))
            items = list(lib.items(args))
            for item in items:
                self.add_item_lyrics(item, ui.should_write())
                if item.lyrics and opts.print:
                    ui.print_(item.lyrics)

            if opts.rest_directory and (
                items := [i for i in items if i.lyrics]
            ):
                RestFiles(Path(opts.rest_directory)).write(items)

        cmd.func = func
        return [cmd]

    def imported(self, _, task: ImportTask) -> None:
        """Import hook for fetching lyrics automatically."""
        if query_str := self.config["auto_ignore"].get():
            query, _ = parse_query_string(query_str, Item)
        else:
            # matches nothing, so all items proceed normally
            query = FalseQuery()

        for item in filterfalse(query.match, task.imported_items()):
            self.add_item_lyrics(item, False)

    def find_lyrics(self, item: Item) -> Lyrics | None:
        """Return the first lyrics match from the configured source search."""
        album, length = item.album, round(item.length)
        matches = (
            self.get_lyrics(a, t, album, length)
            for a, titles in search_pairs(item)
            for t in titles
        )

        return next(filter(None, matches), None)

    def add_item_lyrics(self, item: Item, write: bool) -> None:
        """Fetch and store lyrics for a single item. If ``write``, then the
        lyrics will also be written to the file itself.
        """
        if self.config["local"]:
            return

        if not self.config["force"] and item.lyrics:
            self.info("🔵 Lyrics already present: {}", item)
            return

        existing_lyrics = Lyrics.from_item(item)
        if self.config["keep_synced"] and existing_lyrics.synced:
            self.info("🔵 Keeping synced lyrics: {}", item)
            return

        if new_lyrics := self.find_lyrics(item):
            self.info("🟢 Found lyrics: {}", item)
            if translator := self.translator:
                new_lyrics = translator.translate(new_lyrics, existing_lyrics)

            synced_mode = self.config["synced"].get(bool)
            if synced_mode and existing_lyrics.synced and not new_lyrics.synced:
                self.info(
                    "🔴 Not updating synced lyrics with non-synced ones: {}",
                    item,
                )
                return

            for key in ("backend", "url", "language", "translation_language"):
                item_key = f"lyrics_{key}"
                if value := getattr(new_lyrics, key):
                    item[item_key] = value
                elif item_key in item:
                    del item[item_key]

            lyrics_text = new_lyrics.full_text
        else:
            self.info("🔴 Lyrics not found: {}", item)
            lyrics_text = self.config["fallback"].get()

        if lyrics_text not in {None, item.lyrics}:
            item.lyrics = lyrics_text
            item.store()
            if write:
                item.try_write()

    def get_lyrics(self, artist: str, title: str, *args) -> Lyrics | None:
        """Get first found lyrics, trying each source in turn."""
        self.info("Fetching lyrics for {} - {}", artist, title)
        for backend in self.backends:
            with backend.handle_request():
                if lyrics_info := backend.fetch(artist, title, *args):
                    return lyrics_info

        return None



# This file is part of beets.
# Copyright 2016, Adrian Sampson.
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.

"""Facilities for automatically determining files' correct metadata."""

from __future__ import annotations

from importlib import import_module

# Parts of external interface.
from beets.util.deprecation import deprecate_for_maintainers, deprecate_imports

from .hooks import AlbumInfo, AlbumMatch, TrackInfo, TrackMatch
from .match import Proposal, Recommendation, tag_album, tag_item


def __getattr__(name: str):
    if name == "current_metadata":
        deprecate_for_maintainers(
            f"'beets.autotag.{name}'", "'beets.util.get_most_common_tags'"
        )
        return import_module("beets.util").get_most_common_tags

    return deprecate_imports(
        __name__, {"Distance": "beets.autotag.distance"}, name
    )


__all__ = [
    "AlbumInfo",
    "AlbumMatch",
    "Proposal",
    "Recommendation",
    "TrackInfo",
    "TrackMatch",
    "tag_album",
    "tag_item",
]



# This file is part of beets.
# Copyright 2024, Nicholas Boyd Isacsson.
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.

"""Test the substitute plugin regex functionality."""

from beets.test.helper import PluginTestCase
from beetsplug.substitute import Substitute


class SubstitutePluginTest(PluginTestCase):
    plugin = "substitute"
    preload_plugin = False

    def run_substitute(self, config, cases):
        with self.configure_plugin(config):
            for input, expected in cases:
                assert Substitute().tmpl_substitute(input) == expected

    def test_simple_substitute(self):
        self.run_substitute(
            {
                "a": "x",
                "b": "y",
                "c": "z",
            },
            [("a", "x"), ("b", "y"), ("c", "z")],
        )

    def test_case_insensitivity(self):
        self.run_substitute({"a": "x"}, [("A", "x")])

    def test_unmatched_input_preserved(self):
        self.run_substitute({"a": "x"}, [("c", "c")])

    def test_regex_to_static(self):
        self.run_substitute(
            {".*jimi hendrix.*": "Jimi Hendrix"},
            [("The Jimi Hendrix Experience", "Jimi Hendrix")],
        )

    def test_regex_capture_group(self):
        self.run_substitute(
            {"^(.*?)(,| &| and).*": r"\1"},
            [
                ("King Creosote & Jon Hopkins", "King Creosote"),
                (
                    (
                        "Michael Hurley, The Holy Modal Rounders, Jeffrey"
                        " Frederick & The Clamtones"
                    ),
                    "Michael Hurley",
                ),
                ("James Yorkston and the Athletes", "James Yorkston"),
            ],
        )

    def test_partial_substitution(self):
        self.run_substitute({r"\.": ""}, [("U.N.P.O.C.", "UNPOC")])

    def test_rules_applied_in_definition_order(self):
        self.run_substitute(
            {
                "a": "x",
                "[ab]": "y",
                "b": "z",
            },
            [
                ("a", "x"),
                ("b", "y"),
            ],
        )

    def test_rules_applied_in_sequence(self):
        self.run_substitute(
            {"a": "b", "b": "c", "d": "a"},
            [
                ("a", "c"),
                ("b", "c"),
                ("d", "a"),
            ],
        )



# This file is part of beets.
# Copyright 2016, Adrian Sampson and Diego Moreda.
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.

"""Aid in submitting information to MusicBrainz.

This plugin allows the user to print track information in a format that is
parseable by the MusicBrainz track parser [1]. Programmatic submitting is not
implemented by MusicBrainz yet.

[1] https://wiki.musicbrainz.org/History:How_To_Parse_Track_Listings
"""

import subprocess

from beets import ui
from beets.autotag.match import Recommendation
from beets.plugins import BeetsPlugin
from beets.util import PromptChoice, displayable_path
from beetsplug.info import print_data


class MBSubmitPlugin(BeetsPlugin):
    def __init__(self):
        super().__init__()

        self.config.add(
            {
                "format": "$track. $title - $artist ($length)",
                "threshold": "medium",
                "picard_path": "picard",
            }
        )

        # Validate and store threshold.
        self.threshold = self.config["threshold"].as_choice(
            {
                "none": Recommendation.none,
                "low": Recommendation.low,
                "medium": Recommendation.medium,
                "strong": Recommendation.strong,
            }
        )

        self.register_listener(
            "before_choose_candidate", self.before_choose_candidate_event
        )

    def before_choose_candidate_event(self, session, task):
        if task.rec <= self.threshold:
            return [
                PromptChoice("p", "Print tracks", self.print_tracks),
                PromptChoice("o", "Open files with Picard", self.picard),
            ]

    def picard(self, session, task):
        paths = []
        for p in task.paths:
            paths.append(displayable_path(p))
        try:
            picard_path = self.config["picard_path"].as_str()
            subprocess.Popen([picard_path, *paths])
            self._log.info("launched picard from\n{}", picard_path)
        except OSError as exc:
            self._log.error("Could not open picard, got error:\n{}", exc)

    def print_tracks(self, session, task):
        for i in sorted(task.items, key=lambda i: i.track):
            print_data(None, i, self.config["format"].as_str())

    def commands(self):
        """Add beet UI commands for mbsubmit."""
        mbsubmit_cmd = ui.Subcommand(
            "mbsubmit", help="Submit Tracks to MusicBrainz"
        )

        def func(lib, opts, args):
            items = lib.items(args)
            self._mbsubmit(items)

        mbsubmit_cmd.func = func

        return [mbsubmit_cmd]

    def _mbsubmit(self, items):
        """Print track information to be submitted to MusicBrainz."""
        for i in sorted(items, key=lambda i: i.track):
            print_data(None, i, self.config["format"].as_str())
